(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)

theory Unraveling
imports
  Conditional_Rewriting
  Orthogonality
  Term_Impl
begin

locale unraveling =  
  fixes R :: "('f,'v)ctrs"
  and U :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> ('f,'v)ctxt"
  (* e.g., U ((l,r),s1 = t1, ... sn = tn) i = U_i(\<box>,vars(l) \<union> vars(t1) ... \<union> vars(t i-1) *)
begin

fun lhs_n :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> ('f,'v)term"
  where "lhs_n (lr,cs) 0 = fst lr"
      | "lhs_n (lr,cs) (Suc n) = (U (lr,cs) n) \<langle>snd (cs ! n)\<rangle>"

fun rhs_n :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> ('f,'v)term"
  where "rhs_n (lr,cs) n = (if n < length cs then 
         (U (lr,cs) n)\<langle>fst (cs ! n)\<rangle>
         else snd lr)"

definition rules :: "('f,'v)crule \<Rightarrow> ('f,'v)trs"
  where "rules crule \<equiv> (\<lambda> n. (lhs_n crule n, rhs_n crule n)) ` {n . n \<le> length (snd crule)}"

definition UR :: "('f, 'v) trs" where
  "UR \<equiv> \<Union>(rules ` R)"

lemma UR_simulation_main: assumes lr: "((l, r), cs) \<in> R"
  and i: "i \<le> length cs"
  and cs: "\<And> j. j < i \<Longrightarrow> (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*"
  shows "(l \<cdot> \<sigma>, rhs_n ((l,r),cs) i \<cdot> \<sigma>) \<in> (rstep UR)^+"
proof -
  let ?lr = "((l,r),cs)"
  let ?n = "length cs"
  from lr have lr: "rules ?lr \<subseteq> UR" unfolding UR_def by auto
  hence lr: "\<And> i. i \<le> ?n \<Longrightarrow> (lhs_n ?lr i, rhs_n ?lr i) \<in> UR" unfolding rules_def
    by auto
  show ?thesis
  proof (rule rtrancl_into_trancl2)
    show "(l \<cdot> \<sigma>,rhs_n ?lr 0 \<cdot> \<sigma>) \<in> rstep UR"
      by (rule rstepI[OF lr[of 0], of _ Hole \<sigma>], auto)
    from i cs
    show "(rhs_n ?lr 0 \<cdot> \<sigma>, rhs_n ?lr i \<cdot> \<sigma>) \<in> (rstep UR)^*"
    proof (induct i)
      case 0 thus ?case by simp
    next
      case (Suc i)
      hence i: "i < ?n" and i': "i \<le> ?n" by auto
      let ?f = "U ?lr i"
      let ?U = "\<lambda> t. ?f \<langle> t \<rangle>"
      from Suc(1)[OF i' Suc(3)]
      have "(rhs_n ?lr 0 \<cdot> \<sigma>, rhs_n ?lr i \<cdot> \<sigma>) \<in> (rstep UR)^*" by simp
      also have "rhs_n ?lr i = ?U (fst (cs ! i))"
        using i by simp
      also have "(?U (fst (cs ! i)) \<cdot> \<sigma>, ?U (snd (cs ! i)) \<cdot> \<sigma>) \<in> (rstep UR)^*" 
      proof -
        obtain si ti where csi: "cs ! i = (si,ti)" by force
        from Suc(3)[of i, unfolded csi] have "(si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (rstep UR)^*" by simp
        from ctxt_closedE[OF ctxt_closed_rtrancl[OF ctxt_closed_rstep] this]
        have "\<And> C. (C\<langle>si \<cdot> \<sigma>\<rangle>, C\<langle>ti \<cdot> \<sigma>\<rangle>) \<in> (rstep UR)^*" .
        from this[of "?f \<cdot>c \<sigma>"]
        show ?thesis unfolding csi by simp
      qed
      also have "?U (snd (cs ! i)) = lhs_n ?lr (Suc i)" 
        unfolding lhs_n.simps by simp
      also have "(lhs_n ?lr (Suc i) \<cdot> \<sigma>, rhs_n ?lr (Suc i) \<cdot> \<sigma>) \<in> (rstep UR)"
        by (rule rstepI[OF lr[OF Suc(2)], of _ Hole \<sigma>], auto)
      finally show ?case .
    qed
  qed
qed

lemma SN_UR_quasi_reductive: assumes SN: "SN (rstep UR)"
  shows "quasi_reductive R"
proof -
  let ?S = "(rstep UR)^+"
  let ?NS = "(rstep UR)^*"
  have id: "?S^= = ?NS" by simp
  show ?thesis unfolding quasi_reductive_def
  proof (rule exI[of _ ?S], rule exI[of _ ?NS], unfold id, intro conjI allI impI)
    show "redpair_order ?S ?NS"
    proof
      show "SN ?S" by (rule SN_imp_SN_trancl[OF SN])
      show "ctxt_closed ?NS" by auto
      show "subst_closed ?NS" by (rule subst_closed_rtrancl, auto) 
      show "subst_closed ?S" by (rule subst_closed_trancl, auto) 
    qed auto
    show "ctxt_closed ?S" by auto
  next
    fix l r cs \<sigma>
    assume lr: "((l, r), cs) \<in> R"
      and cs: "\<forall>j<length cs. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*"
    from UR_simulation_main[OF lr, of "length cs" \<sigma>] cs
    show "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> (rstep UR)^+" by simp
  next
    fix l r cs \<sigma> i
    assume lr: "((l, r), cs) \<in> R"
    and i: "i < length cs"
    and cs: "\<forall>j<i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*"
    from UR_simulation_main[OF lr, of i \<sigma>] i cs
    have lm: "(l \<cdot> \<sigma>, rhs_n ((l, r), cs) i \<cdot> \<sigma>) \<in> (rstep UR)^+" (is "(?l,?m) \<in> _") by simp
    have mr: "?m \<unrhd> fst (cs ! i) \<cdot> \<sigma>" (is "_ \<unrhd> ?r")
      by (rule supteq_stable, insert i, auto)
    let ?R = "(rstep UR)^+ \<union> {\<rhd>}"
    from mr have "(?m,?r) \<in> ({\<rhd>})^=" by auto
    hence mr: "(?m,?r) \<in> ?R^*" by auto
    have lm: "(?l,?m) \<in> ?R^+"
      by (rule trancl_mono[OF lm], auto)
    from lm mr show lr: "(?l,?r) \<in> ?R^+" by auto
  qed
qed

lemma UR_complete: assumes step: "(s,t) \<in> cstep R"
  shows "(s,t) \<in> (rstep UR)^+"
proof -
  from step[unfolded cstep_def] obtain n where "(s,t) \<in> cstep_n R n" ..
  thus ?thesis
  proof (induct n arbitrary: s t)
    case 0 thus ?case by simp
  next
    case (Suc n) note IH = this
    from IH(2) obtain l r cs C \<sigma> where lr: "((l,r),cs) \<in> R" and
    cs: "\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cstep_n R n)^*"
    and s: "s = C \<langle> l \<cdot> \<sigma> \<rangle>" and t: "t = C \<langle> r \<cdot> \<sigma> \<rangle>" by (blast elim: cstep_n_SucE)
    let ?lr = "((l,r),cs)"
    let ?n = "length cs"
    have rsig: "r \<cdot> \<sigma> = rhs_n ?lr ?n \<cdot> \<sigma>" by auto
    have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> (rstep UR)^+" unfolding rsig
    proof (rule UR_simulation_main[OF lr, of "length cs" \<sigma>])
      fix j
      assume j: "j < length cs"
      obtain sj tj where csj: "cs ! j = (sj,tj)" by force
      from j have "(sj,tj) \<in> set cs" unfolding csj[symmetric] set_conv_nth by auto
      from cs[OF this] have steps: "(sj \<cdot> \<sigma>, tj \<cdot> \<sigma>) \<in> (cstep_n R n)^*" .
      {
        fix s t
        assume "(s,t) \<in> (cstep_n R n)^*"
        hence "(s,t) \<in> (rstep UR)^*"
        proof (induct)
          case base thus ?case by auto
        next
          case (step t u)
          from step(3) IH(1)[OF step(2)] show ?case by auto
        qed
      }
      from this[OF steps]
      show "(fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*" 
        unfolding csj by auto
    qed simp
    from ctxt_closedE[OF ctxt_closed_trancl[OF ctxt_closed_rstep] this, of C]
    show ?case unfolding s t .
  qed
qed

end

definition "eqs \<mu> = map (\<lambda>(x, t). (Var x, t)) \<mu>"

hide_const (open) eqs

definition funs_ctrs :: "('f,'v)ctrs \<Rightarrow> 'f set" where
 "funs_ctrs R  \<equiv> \<Union>((\<lambda> \<rho>. funs_rule (fst \<rho>) \<union> funs_trs (set (snd \<rho>))) ` R)"

definition vars_crule :: "('f,'v)crule \<Rightarrow> 'v set" where
 "vars_crule \<rho> \<equiv> vars_rule (fst \<rho>) \<union> vars_trs (set (snd \<rho>))"

definition vars_ctrs :: "('f,'v)ctrs \<Rightarrow> 'v set" where
 "vars_ctrs R  \<equiv> \<Union>((\<lambda> \<rho>. vars_rule (fst \<rho>) \<union> vars_trs (set (snd \<rho>))) ` R)"

context unraveling
begin

abbreviation ss :: "('f,'v) crule \<Rightarrow> ('f,'v) term set" where "ss \<rho> \<equiv> lhss (set (snd \<rho>))"
abbreviation ts :: "('f,'v) crule \<Rightarrow> ('f,'v) term set" where "ts \<rho> \<equiv> rhss (set (snd \<rho>))"
abbreviation clhs ::  "('f,'v) crule \<Rightarrow> ('f,'v) term" where "clhs \<rho> \<equiv> fst (fst \<rho>)"
abbreviation crhs ::  "('f,'v) crule \<Rightarrow> ('f,'v) term" where "crhs \<rho> \<equiv> snd (fst \<rho>)"

(* start indexing conditions from 0 *)
(* predicate for valid X variable enumeration *)
definition X_vars :: "(('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> bool" where
 "X_vars X = (\<forall> \<rho> \<in> R. \<forall> i < length (snd \<rho>). (
  set (X \<rho> i) = vars_term (clhs \<rho>) \<union> \<Union>(vars_term ` (rhss (set (List.take i (snd \<rho>)))))) \<and> 
   distinct (X \<rho> i))"

(* predicate for valid Y variable enumeration *)
definition Y_vars :: "(('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> bool" where
 "Y_vars Y = (\<forall> \<rho> \<in> R. \<forall> i < length (snd \<rho>). (
  set (Y \<rho> i) = vars_term (snd (fst \<rho>)) \<union> vars_term (snd (snd \<rho>!i)) \<union> 
    (vars_trs (set (List.drop i (snd \<rho>)) ))) \<and> distinct (Y \<rho> i))"

definition Z_vars :: "(('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> (('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> (('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> bool" where
 "Z_vars Z X Y = (X_vars X \<and> Y_vars Y \<and> (\<forall> \<rho> \<in> R. \<forall> i < length (snd \<rho>).
   set (Z \<rho> i) =  set (X \<rho> i) \<inter> set (Y \<rho> i) \<and> distinct (Z \<rho> i)))"

abbreviation ultra :: "(('f,'v) trs \<Rightarrow> bool) \<Rightarrow> bool" where "ultra P \<equiv> P (UR)"

abbreviation ultra_rule :: "(('f,'v) trs \<Rightarrow> bool) \<Rightarrow> ('f,'v) crule \<Rightarrow> bool" 
 where "ultra_rule P \<rho> \<equiv> P (rules \<rho>)"

 
 
end

(* locale demanding conditions for U_opt *)
locale unraveling_opt = unraveling R U 
  for R :: "('f,'v) ctrs" and U :: "('f,'v) crule \<Rightarrow> nat => ('f,'v) ctxt" +
  fixes X :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list"
  fixes Y :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list"
  fixes Z :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list"
  fixes fresh :: "nat \<Rightarrow> 'v list "
  assumes U_cond:"\<forall> \<rho> \<in> R. \<forall>n < length (snd \<rho>). \<exists> f. 
   (U \<rho> n = (More f Nil Hole (map Var (Z \<rho> n))) \<and> f \<notin> funs_ctrs R \<and>
   \<not>(\<exists> \<rho>' n' b c a. \<rho>' \<in> R \<and> n' < length (snd \<rho>') \<and> (\<rho> \<noteq> \<rho>' \<or> n \<noteq> n') \<and> 
     U \<rho>' n' = More f b c a))"
  and Z:"Z_vars Z X Y"
  and dctrs:"\<forall> \<rho> \<in> R. \<forall> i < length (snd \<rho>). 
   vars_term (fst ((snd \<rho>)!i)) \<subseteq> (vars_term (clhs \<rho>) \<union> \<Union> ((\<lambda> j. vars_term (snd ((snd \<rho>)!j))) ` {j. j < i}))"
  and type3:"\<forall> \<rho> \<in> R. vars_term (snd (fst \<rho>)) \<subseteq> (vars_term (clhs \<rho>) \<union> \<Union> ((\<lambda> j. vars_term (snd ((snd \<rho>) ! j))) ` {j. j < length (snd \<rho>)}))"
  and fresh_vars:"\<forall> n. length (fresh n) = n \<and> distinct (fresh n)"
begin


(* some properties of X, Y, Z variables *)
lemma X_vars_Suc :
 assumes inR:"\<rho> \<in> R" and il:"Suc i < length (snd \<rho>)" 
 shows "set (X \<rho> (Suc i)) = set (X \<rho> i) \<union> vars_term (snd (snd \<rho> ! i))"
proof-
 let ?v = "\<lambda> m. \<Union>(vars_term ` rhss (set (take m (snd \<rho>))))"
 from il have il':"i < length (snd \<rho>)" by auto 
 from il take_Suc_conv_app_nth[OF il'] have "set (take (Suc i) (snd \<rho>)) = set (take i (snd \<rho>)) \<union> {snd \<rho> ! i}" by force
 hence "rhss (set (take (Suc i) (snd \<rho>))) = rhss(set (take i (snd \<rho>))) \<union> {snd (snd \<rho> ! i)}" by blast
 hence vs:"?v (Suc i) = ?v i \<union> vars_term (snd (snd \<rho> ! i))" by force
 from il il' conjunct1[OF Z[unfolded Z_vars_def X_vars_def], rule_format, OF inR il]
      conjunct1[OF Z[unfolded Z_vars_def X_vars_def], rule_format, OF inR il'] show ?thesis unfolding vs by auto
qed

lemma X_vars_mono :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and ji:"j \<le> i"
 shows "set (X \<rho> j) \<subseteq> set (X \<rho> i)"
proof(cases "j = i", simp)
 case False
 with ji have ji:"j < i" by auto
 let ?v = "\<lambda> m. \<Union>(vars_term ` rhss (set (take m (snd \<rho>))))"
 from Z[unfolded Z_vars_def X_vars_def] il inR have
  si:"set (X \<rho> i) = vars_term (clhs \<rho>) \<union> ?v i" by auto
 from Z[unfolded Z_vars_def X_vars_def] xt1(10)[OF il ji] inR have
  sj:"set (X \<rho> j) = vars_term (clhs \<rho>) \<union> ?v j" by meson
 have "?v j \<subseteq> ?v i" by (rule Union_mono, rule image_mono, rule image_mono, 
  simp add:set_take_subset_set_take[OF less_imp_le[OF ji]]) 
 from Un_mono[OF subset_refl this] show ?thesis unfolding si sj by auto
qed

lemma l_vars_X_vars :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" shows "vars_term (clhs \<rho>) \<subseteq> set (X \<rho> i)" 
using Z[unfolded Z_vars_def X_vars_def] il inR by blast

lemma l_vars_X0_vars :
 assumes inR:"\<rho> \<in> R" and il:"0 < length (snd \<rho>)" shows "vars_term (clhs \<rho>) = set (X \<rho> 0)" 
using Z[unfolded Z_vars_def X_vars_def] il inR by simp


lemma Y_vars_mono :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and ji:"j \<le> i"
 shows "set (Y \<rho> j) \<supseteq> set (Y \<rho> i)"
proof(cases "j = i", simp)
 case False
 with ji have ji:"j < i" by auto
 from ji obtain k where i:"i = j + k " 
by (metis add_Suc_right less_iff_Suc_add)
 let ?v = "\<lambda> m. (vars_trs (set (List.drop m (snd \<rho>)) ))"
 let ?r = "vars_term (snd (fst \<rho>))"
 let ?ti = "\<lambda> k. vars_term (snd (snd \<rho>!k))"
 from Z[unfolded Z_vars_def Y_vars_def] il inR have
  si:"set (Y \<rho> i) = ?r \<union> ?ti i \<union> ?v i" by auto
 from Z[unfolded Z_vars_def Y_vars_def] xt1(10)[OF il ji] inR have
  sj:"set (Y \<rho> j) = ?r \<union> ?ti j \<union> ?v j" by meson 
 from set_drop_subset_set_drop[OF less_imp_le[OF ji]] 
  have 1:"?v i \<subseteq> ?v j" unfolding vars_trs_def Union_image_eq[symmetric] by fast
 from il[unfolded i] have "k < length (drop j (snd \<rho>))" unfolding length_drop by linarith
 from nth_drop[OF less_imp_le[OF il[unfolded i]]] nth_mem[OF this] 
  have "(snd \<rho> ! i) \<in> set (drop j (snd \<rho>))" unfolding i by auto
 hence "vars_term (snd (snd \<rho> ! i)) \<subseteq> vars_trs (set (drop j (snd \<rho>)))" 
  unfolding vars_trs_def Union_image_eq[symmetric] using vars_rule_def by (metis Sup_upper Un_subset_iff imageI)
 from Un_least[OF this 1] Un_upper2 have "?ti i \<union> ?v i \<subseteq> ?ti j \<union> ?v j" by auto
 from Un_mono[OF subset_refl this] show ?thesis unfolding si sj by auto
qed

lemma Z_imp_X :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and z:"x \<in> set (Z \<rho> i)" 
 shows "x \<in> set (X \<rho> i)"
 using conjunct2[OF conjunct2[OF Z[unfolded Z_vars_def]], rule_format, OF inR il] z by blast

lemma Z_imp_Y :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and z:"x \<in> set (Z \<rho> i)" 
 shows "x \<in> set (Y \<rho> i)"
 using conjunct2[OF conjunct2[OF Z[unfolded Z_vars_def]], rule_format, OF inR il] z by blast

lemma X_Y_imp_Z :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and x:"x \<in> set (X \<rho> i)" and y:"x \<in> set (Y \<rho> i)"
 shows "x \<in> set (Z \<rho> i)"
 using  conjunct2[OF conjunct2[OF Z[unfolded Z_vars_def]], rule_format, OF inR il] x y by blast


lemma s_imp_Y :
 assumes inR:"\<rho> \<in> R" and il:"Suc i < length (snd \<rho>)" and x:"x \<in> vars_term (fst (snd \<rho> ! (Suc i)))"
 shows "x \<in> set (Y \<rho> i)"
proof-
 from Z[unfolded Z_vars_def] have y:"Y_vars Y" by auto 
 from il have len:"1 < length (drop i (snd \<rho>))" by fastforce 
 from nth_drop[OF less_imp_le[OF il[unfolded Suc_eq_plus1]]] nth_mem[OF len]
  have aux: "(snd \<rho> ! (Suc i)) \<in> (set (drop i (snd \<rho>)))" by auto
 from x have xr:"x \<in> vars_rule (snd \<rho> ! Suc i)" unfolding vars_rule_def by blast
 have "x \<in> vars_trs (set (drop i (snd \<rho>)))" unfolding vars_trs_def Union_image_eq[symmetric] 
  by(rule UnionI[OF _ xr], insert aux, auto)
 with y[unfolded Y_vars_def, rule_format, OF inR less_trans[OF lessI il]] show ?thesis by blast
qed

lemma X_le_Z_imp_Z :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and ji:"j \<le> i" and 
 x:"x \<in> set (X \<rho> j)" and z:"x \<in> set (Z \<rho> i)"
 shows "x \<in> set (Z \<rho> j)"
proof(cases "i=j")
 case True with z show ?thesis by auto
 next
 case False
 from z Z_imp_Y[OF inR il z] ji inR il have "x \<in> set (Y \<rho> i)" by blast
 with Y_vars_mono[OF inR il, unfolded order.strict_iff_order] ji False  have "x \<in> set (Y \<rho> j)" by fast
 with x show ?thesis using conjunct2[OF conjunct2[OF Z[unfolded Z_vars_def]], rule_format, OF inR] 
  order.strict_trans1[OF ji il]  by blast
qed

lemma Z_interval :
 assumes inR:"\<rho> \<in> R" and kl:"k < length (snd \<rho>)" and ji:"j \<le> i" and ik:"i \<le> k" and 
 xj:"x \<in> set (Z \<rho> j)" and xk:"x \<in> set (Z \<rho> k)"
 shows "x \<in> set (Z \<rho> i)"
proof-
  from ji kl ik have jl:"j < length (snd \<rho>)" by auto
  from kl ik have il:"i < length (snd \<rho>)" by auto
  from X_vars_mono[OF inR il ji] Z_imp_X[OF inR jl xj] have x:"x \<in> set (X \<rho> i)" by blast
  from Y_vars_mono[OF inR kl ik] Z_imp_Y[OF inR kl xk] have "x \<in> set (Y \<rho> i)" by blast
  from X_Y_imp_Z[OF inR il x this] show ?thesis by fast
qed

(* Theorem 3.9(1), => in NSS12 *)
lemma L3_9a :
 assumes inR:"\<rho> \<in> R" and ull:"ultra_rule left_linear_trs \<rho>"
 shows "linear_term (clhs \<rho>) \<and> (\<forall> i < length (snd \<rho>). linear_term ( snd(snd \<rho> ! i))) \<and>
      (\<forall> i < length (snd \<rho>). vars_term (snd ((snd \<rho>) ! i)) \<inter> set (X \<rho> i) = {})"
proof-
 obtain l r cs where rho:"\<rho> = ((l,r),cs)" using PairE by metis
 with ull have ll:"\<And> l' r'. (l',r') \<in> rules \<rho> \<Longrightarrow> linear_term l'" unfolding left_linear_trs_def by blast
 hence "linear_term (lhs_n \<rho> 0)"  unfolding rules_def by fast
 hence 1:"linear_term (clhs \<rho>)" unfolding rules_def using lhs_n.simps(1) prod.collapse by metis
 { fix i
  assume i:"i < length (snd \<rho>)"
  hence i':"Suc i \<le> length (snd \<rho>)" by auto
  let ?ti = "snd (snd \<rho> ! i)"
  from  ll[unfolded rules_def] i' have lin:"linear_term (lhs_n \<rho> (Suc i))" by blast
  from U_cond[rule_format, OF inR i] obtain f where "U \<rho> i = More f [] \<box> (map Var (Z \<rho> i))" by fast
  from lin[unfolded rho lhs_n.simps(2) this[unfolded rho] ctxt_apply.simps] rho have 
   "linear_term (Fun f ((snd (cs!i)) # (map Var (Z \<rho> i))))" by auto
  from this[unfolded linear_term.simps(2)] have 2:"linear_term ?ti" unfolding rho by force
  have "vars_term (snd ((snd \<rho>)!i)) \<inter> set (X \<rho> i) = {}"
   proof(rule ccontr)
    assume ne:"vars_term (snd (snd \<rho> ! i)) \<inter> set (X \<rho> i) \<noteq> {}"
    let ?ti = "(snd (snd \<rho> ! i))"
    let ?zs = "Z \<rho> i"
    from Z[unfolded Z_vars_def Y_vars_def] have "vars_term ?ti \<subseteq> set (Y \<rho> i)" using inR i by force
    with Z[unfolded Z_vars_def] ne have iz:"vars_term ?ti \<inter> set (Z \<rho> i) \<noteq> {}" using inR i by fast
    have zs:"\<Union> (set (map vars_term (map Var ?zs))) = set ?zs" unfolding map_map vars_term_Var_id by force
    from iz have p:"\<not> (is_partition (map vars_term (?ti # (map Var ?zs))))" 
     unfolding list.simps(9) is_partition_Cons zs by blast
    let ?t = "(U \<rho> i)\<langle>snd (snd \<rho> ! i)\<rangle>"
    from U_cond[rule_format, OF inR i] obtain f where Uri:"U \<rho> i = More f [] \<box> (map Var (Z \<rho> i))" by fast
    from p have "\<not> linear_term ?t" unfolding Uri by simp
    hence l:"\<not> linear_term (lhs_n \<rho> (Suc i))" unfolding rho by auto
    have "linear_term (lhs_n \<rho> (Suc i))" by(rule ll[unfolded rules_def], insert i, auto)
    with l show False by auto
   qed 
 with 2 have "linear_term ?ti \<and> vars_term (snd ((snd \<rho>)!i)) \<inter> set (X \<rho> i) = {}" by auto
 }
 with 1 show ?thesis by auto
qed

abbreviation si :: "('f,'v) crule \<Rightarrow> nat =>  ('f,'v) term" where "si \<rho> i \<equiv> fst ((snd \<rho>)!i)"
abbreviation ti :: "('f,'v) crule \<Rightarrow> nat =>  ('f,'v) term" where "ti \<rho> i \<equiv> snd ((snd \<rho>)!i)"



lemma part_rev : assumes p:"is_partition xs" shows "is_partition (rev xs)"
 unfolding is_partition_def length_rev 
proof(rule,rule)
 fix j
 assume j:"j < length xs"
 show "\<forall>i<j. rev xs ! i \<inter> rev xs ! j = {}" proof(rule,rule)
  fix i
  assume i:"i < j"
  from j have 1:"length xs  - (Suc i) < length xs" by fastforce
  from i j have 2:"length xs  - (Suc j) < length xs  - (Suc i)"
    by (metis Suc_less_eq diff_less_mono2 less_trans_Suc)
  from p[unfolded is_partition_def] 1 2 
  show "rev xs ! i \<inter> rev xs ! j  = {}" unfolding rev_nth[OF j] rev_nth[OF order.strict_trans[OF i j]] by blast
 qed
qed

lemma is_partition_rev : "is_partition (rev xs) = is_partition xs "
 using part_rev rev_rev_ident by force

(* replace by all_ctxt_closed_subst_step? *)
lemma substs_rsteps':
  shows "(\<And>x. x \<in> vars_term t \<Longrightarrow> (\<sigma> x, \<tau> x) \<in> (cstep R)\<^sup>*) \<Longrightarrow> (t \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (cstep R)\<^sup>*"
proof (induct t)
  case (Var y) 
  thus ?case using assms [of y] by simp
next
  case (Fun f ts)
  have "\<And> s. s \<in> set ts \<Longrightarrow>  vars_term s \<subseteq> vars_term (Fun f ts)" by auto
  with Fun have "\<And> s. s \<in> set ts \<Longrightarrow> (s \<cdot> \<sigma>, s \<cdot> \<tau>) \<in> (cstep R)\<^sup>*" by blast
  then have "\<forall>i<length (map (\<lambda>t. t \<cdot> \<sigma>) ts).
    (map (\<lambda>t. t \<cdot> \<sigma> ) ts ! i, map (\<lambda>t. t \<cdot> \<tau>) ts ! i) \<in> (cstep R)\<^sup>*" by auto
  from args_steps_imp_steps[OF cstep_ctxt_closed _ this] show ?case by simp
qed

(* Lemma A.1 in NSS12 *)
(* partition of variable sets in substitution \<sigma> (different from paper *)
definition X_part :: "('f,'v) crule \<Rightarrow> 'v set list" where
 "X_part \<rho> \<equiv> ((set (X \<rho> 0)) # (map (\<lambda>i. set (X \<rho> (Suc i)) -  (set (X \<rho> i))) [0..< (length (snd \<rho>) - 1)]) @
             [(vars_term (crhs \<rho>) \<union> (vars_term (ti \<rho> (length (snd \<rho>) - 1)))) - (set (X \<rho> (length (snd \<rho>) - 1)))])"


lemma X_part: assumes inR:"\<rho> \<in> R" 
 and l:"length (snd \<rho>) \<noteq> 0 "
 shows "is_partition (X_part \<rho>)"
proof-
 let ?k = "length (snd \<rho>)"
 from not0_implies_Suc[OF l] obtain k' where k_suc:"?k = Suc k'" by auto
 let ?xk = "(vars_term (crhs \<rho>) \<union> (vars_term (ti \<rho> k'))) - (set (X \<rho> k'))"
 let ?xs = "\<lambda> m. (set (X \<rho> 0)) # (map (\<lambda>i. set (X \<rho> (Suc i)) -  (set (X \<rho> i))) [0..<m])"
 fix j
 have key:"\<forall>j. j < ?k \<longrightarrow> (is_partition (rev (?xs j)) \<and> \<Union>set (?xs j) \<subseteq> set (X \<rho> j))" proof(rule allI, rule)
  fix j 
  assume "j < ?k"
  thus "is_partition (rev (?xs j))  \<and> \<Union>set (?xs j) \<subseteq> set (X \<rho> j)" proof(induct j)
  case 0 
   have s:"?xs 0 = [set (X \<rho> 0)]" by auto
   thus ?case unfolding s using rev.simps is_partition_Cons is_partition_Nil by force
  next
  case (Suc j)
   assume lt:"Suc j < ?k"
   with Suc have p:"is_partition (rev (?xs j))" and ss:"\<Union>set (?xs j) \<subseteq> set (X \<rho> j)" by auto
   let ?xi = "\<lambda> j. set (X \<rho> (Suc j)) -  (set (X \<rho> j))"
   have "?xs (Suc j) = (?xs j) @ [?xi j]" unfolding upt_Suc_append[OF le0] map_append by fastforce
   hence r:"rev (?xs (Suc j)) =  (?xi j) #  (rev (?xs j))" using rev_eq_Cons_iff[of "?xs (Suc j)"] by auto
   have "(?xi j) \<inter> set (X \<rho> j) = {}" by auto
   with ss have "(?xi j) \<inter> \<Union>set (rev (?xs j)) = {}" by auto
   with Suc r have p:"is_partition (rev (?xs (Suc j)))" using is_partition_Cons[of "?xi j" "rev (?xs j)"] by fastforce
   from ss X_vars_mono[OF inR lt less_imp_le[OF lessI]] 
     have "\<Union>set (?xs (Suc j)) \<subseteq> set (X \<rho> (Suc j))" 
     unfolding  upt_Suc_append[OF le0] map_append list.set(2) set_append by auto
   with p show ?case by auto
  qed
  qed
 have lt:"k' < ?k" using k_suc by auto
 from key this have p:"is_partition (rev (?xs k'))" "\<Union>set (?xs k') \<subseteq> set (X \<rho> k')" unfolding X_part_def by auto
 from p(2) have isc:"?xk \<inter> \<Union>set (rev (?xs k')) = {}" by auto
 from k_suc have kk:"?k - 1 = k'" by auto
 have aux:"rev (?xs k' @ [?xk]) = ?xk # rev (?xs k')"  using rev_eq_Cons_iff by fastforce
 with p isc have "is_partition (rev (?xs k' @ [?xk]))" unfolding aux is_partition_Cons by blast
 thus ?thesis using is_partition_rev unfolding X_part_def kk by force
qed

 
lemma Z_part: assumes inR:"\<rho> \<in> R" 
 and i:"i < length (snd \<rho>)"
 shows "is_partition (map (\<lambda> z. {z}) (Z \<rho> i))"
proof-
 from Z[unfolded Z_vars_def] inR i have "distinct (Z \<rho> i)" by fast
 with distinct_conv_nth show ?thesis unfolding is_partition_def by force
qed

lemma Z_part': assumes inR:"\<rho> \<in> R" 
 and i:"i < length (snd \<rho>)"
 and ull:"ultra_rule left_linear_trs \<rho>"
 shows "is_partition (vars_term (ti \<rho> i) # map (\<lambda> z. {z}) (Z \<rho> i))"
proof-
 from Z_part[OF inR i] have p:"is_partition (map (\<lambda> z. {z}) (Z \<rho> i))" by auto
 from L3_9a[OF inR ull] i have "vars_term (ti \<rho> i) \<inter> set (X \<rho> i) = {}" by auto
 with Z[unfolded Z_vars_def] inR i have "vars_term (ti \<rho> i) \<inter> set (Z \<rho> i) = {}" by auto
 with p is_partition_Cons show ?thesis by fastforce
qed

lemma X_part_nth :
 assumes inR:"\<rho> \<in> R" 
 and l:"length (snd \<rho>) \<noteq> 0 "
 and jk:"Suc j < length (snd \<rho>)"
 shows " (X_part \<rho>) ! (Suc j) =  set (X \<rho> (Suc j)) - set (X \<rho> j)"
proof-
 let ?k = "length (snd \<rho>)"
 let ?f = "\<lambda>i. set (X \<rho> (Suc i)) - set (X \<rho> i)"
 from jk have x:"j < ?k - 1 - 0" by auto
 hence "map ?f [0..<?k - 1] ! j = set (X \<rho> (Suc j)) - set (X \<rho> j)" 
  unfolding nth_map[of j "[0..<?k - 1]", unfolded length_upt, OF x] 
  using nth_upt by force
 with nth_map[of j "[0..<?k - 1]", unfolded length_upt, OF x] show ?thesis unfolding X_part_def nth_Cons_Suc 
   nth_append[of "map ?f [0..<?k - 1]" _ j, unfolded length_map length_upt eqTrueI[OF x] if_True] by blast
qed

lemma X_part_last :
 assumes inR:"\<rho> \<in> R" 
 and l:"length (snd \<rho>) \<noteq> 0 "
 shows " (X_part \<rho>) ! (length (snd \<rho>)) = 
  (vars_term (crhs \<rho>) \<union> (vars_term (ti \<rho> (length (snd \<rho>) - 1)))) - (set (X \<rho> (length (snd \<rho>) - 1)))"
proof-
 let ?k = "length (snd \<rho>)" 
 let ?f = "\<lambda>i. set (X \<rho> (Suc i)) - set (X \<rho> i)"
 from l have kl:"?k = length ((set (X \<rho> 0)) # map ?f [0..<?k - 1])" unfolding length_Cons length_map length_upt by simp
 with nth_append_length show ?thesis unfolding X_part_def append_Cons[symmetric] by (metis (no_types, lifting))
qed

lemma A1:
 assumes inR:"\<rho> \<in> R" 
 and ull:"ultra_rule left_linear_trs \<rho>"
 and l:"length (snd \<rho>) \<noteq> 0 "
 and sl:"length (\<sigma> :: ('f,'v) subst list) = Suc (length (snd \<rho>))"
 and st:"\<forall> i < length (snd \<rho>). ((si \<rho> i) \<cdot> (\<sigma>!i), (ti \<rho> i) \<cdot> (\<sigma>!Suc i) ) \<in> (cstep R)^*"
 and zrew:"\<forall> i < length (snd \<rho>) . \<forall> z \<in> set(Z \<rho> i) . (Var z \<cdot> (\<sigma>!i), Var z \<cdot> (\<sigma>!Suc i)) \<in> (cstep R)^*"
 shows
 "(clhs \<rho> \<cdot> (\<sigma>!0), crhs \<rho> \<cdot> (\<sigma>!(length (snd \<rho>)))) \<in> (cstep R)^+"
proof-
 let ?k = "length (snd \<rho>)"
 let ?xs = "X_part \<rho>"
 let ?\<sigma> = "fun_merge \<sigma> ?xs"
 from not0_implies_Suc l obtain k' where k_suc:"?k = Suc k'" by auto
 have lx:"Suc ?k = length ?xs" unfolding X_part_def using k_suc by fastforce
 from X_part[OF inR l] have p:"is_partition ?xs" by fast
 { fix t :: "('f,'v) term" fix i
   assume x:"i < Suc ?k" and y:"vars_term t \<subseteq> ?xs!i"
   from y fun_merge_part[OF p] x[unfolded lx] have "\<And> x. x \<in> (vars_term t) \<Longrightarrow> ?\<sigma> x = (\<sigma>!i) x" by blast
   from term_subst_eq this have "t \<cdot> ?\<sigma> = t \<cdot> (\<sigma>!i)" by blast
 } note subst_fact = this
 let ?l = "clhs \<rho>" let ?r = "crhs \<rho>" let ?cs = "snd \<rho>"
 from inR have rho:"((?l,?r),?cs) \<in> R " by auto
 have "vars_term ?l \<subseteq> ?xs ! 0" using l_vars_X_vars[OF inR, OF l[unfolded neq0_conv]] unfolding X_part_def by auto
 from subst_fact[OF _ this] lx have lsigma:"?l \<cdot> ?\<sigma> = ?l \<cdot> (\<sigma>!0)" by fast
 (* next, show t_i \<sigma>_{i+1} = t_i \<sigma> *)
  { fix i
    assume ik:"i < ?k"  
    from conjunct2[OF conjunct2[OF L3_9a[OF inR ull]], rule_format,OF ik] have ss':"vars_term (ti \<rho> i) \<inter> set (X \<rho> i) = {}" by auto
    have "vars_term (ti \<rho> i) \<subseteq> ?xs!Suc i"  proof(cases "Suc i < ?k")
    case True
     from ik min_less_iff_conj nth_mem[of i "take (Suc i)(snd \<rho>)", unfolded length_take nth_take[OF lessI]] 
      have "ti \<rho> i \<in> rhss (set (take (Suc i)(snd \<rho>)))" by blast
     with subsetI image_eqI[OF refl] have "vars_term (ti \<rho> i) \<subseteq> \<Union>(vars_term ` rhss (set (take (Suc i) (snd \<rho>))))" by blast
     hence ss:"vars_term (ti \<rho> i) \<subseteq> set (X \<rho> (Suc i))" 
      using conjunct1[OF Z[unfolded Z_vars_def], unfolded X_vars_def,rule_format,OF inR True] by auto
     from X_part_nth[OF inR _ True] k_suc have "?xs!Suc i = set (X \<rho> (Suc i)) - set (X \<rho> i)" by auto
     with ss ss' show "vars_term (ti \<rho> i) \<subseteq> ?xs!Suc i" by blast
    next
    case False
     with ik have i:"i = ?k - 1" and ii: "Suc i = ?k" by auto
     have xk: "?xs ! (Suc i) = vars_term (crhs \<rho>) \<union> vars_term (ti \<rho> i) - set (X \<rho> i)" 
      unfolding X_part_def i k_suc
      using nth_append_length[of "set (X \<rho> 0) # map (\<lambda>i. set (X \<rho> (Suc i)) - set (X \<rho> i)) [0..<Suc k' - 1]", unfolded length_Cons length_map length_upt]
      unfolding length_Cons length_map length_upt by force
     with ss' show "vars_term (ti \<rho> i) \<subseteq> ?xs!Suc i" unfolding ii by blast
    qed
    from subst_fact[OF _ this] ik have "(ti \<rho> i) \<cdot> (\<sigma>! (Suc i)) = (ti \<rho> i) \<cdot> ?\<sigma>" by fastforce
  } note t_fact = this 
(* attempt of more general auxiliary result for s_i and r case *)
{ fix j x
  assume y:"x \<in> set (Y \<rho> j)" 
   and lhs_or_ti:"x \<in> vars_term ?l \<or> (\<exists> k < Suc j. x \<in> vars_term (ti \<rho> k))"
   and ik:"Suc j \<le> ?k"
  from y Y_vars_mono[OF inR] Suc_le_lessD[OF ik] have y:"\<And> j'. j' \<le> j \<Longrightarrow> x \<in> set (Y \<rho> j')" by blast
  have "(?\<sigma> x, (\<sigma>! (Suc j)) x) \<in> (cstep R)^*" 
   proof(cases "x \<in> vars_term ?l")
    case True
     with l_vars_X_vars[OF inR] k_suc have "x \<in> ?xs!0"  unfolding X_part_def nth_Cons_0 by auto
     from fun_merge_part[OF p _ this, unfolded lx[symmetric] k_suc] have  sx0:"?\<sigma> x = (\<sigma>!0) x" by auto
       from l_vars_X_vars[OF inR] True k_suc X_vars_mono[OF inR _ le0] 
         order.strict_trans2[OF _ ik] have x:"\<And> j'. j' < Suc j \<Longrightarrow> x \<in> set (X \<rho> j')" by blast
       from x y X_Y_imp_Z[OF inR] ik k_suc have z:"\<And> j'. j' < Suc j \<Longrightarrow> x \<in> set (Z \<rho> j')" by simp
       hence xZ0:"x \<in> set (Z \<rho> 0)" by auto
       from z ik have "((\<sigma> ! 0) x,  (\<sigma> ! (Suc j)) x) \<in> (cstep R)\<^sup>*" proof (induct j)
        case 0 from zrew[rule_format, of 0] k_suc xZ0 show ?case by simp
        next
        case (Suc k) 
         hence 1:"Suc k < ?k" and 2:"\<And> j. j < Suc k \<Longrightarrow> x \<in> set (Z \<rho> j)" by auto
         from zrew[rule_format, OF 1 Suc(2)] Suc(1) 2 1 show ?case by fastforce
        qed
       with sx0 show ?thesis by auto
      next 
      case False
       with lhs_or_ti obtain j' where ji:"j' < Suc j" and xj:"x \<in> vars_term (ti \<rho> j')" by auto
       from ji ik have 1:"Suc j' \<le> ?k" by auto
       from  L3_9a[OF inR ull] order.strict_trans2[OF ji ik] have x:"vars_term (ti \<rho> j') \<inter> set (X \<rho> j') = {}" by fast
       from nth_take[OF lessI] nth_mem[of j' "take (Suc j') (snd \<rho>)", unfolded nth_take[OF lessI] length_take min_less_iff_conj]  
        order.strict_trans2[OF ji ik] image_iff[of "ti \<rho> j'" snd] lessI[of j']
        have "ti \<rho> j' \<in>  rhss (set (take (Suc j') ?cs))" by blast
       with xj image_iff have xu:"x \<in> \<Union>(vars_term ` rhss (set (take (Suc j') ?cs)))" unfolding Union_iff by blast
       from xu conjunct1[OF Z[unfolded Z_vars_def], unfolded X_vars_def, rule_format, OF inR] False
          have xsj: "Suc j' < ?k \<Longrightarrow> x \<in> set (X \<rho> (Suc j'))" by blast
       have "x \<in> ?xs!(Suc j')"
        proof(cases "Suc j' < ?k")
         case True
          from xsj[OF True] x xj X_part_nth[OF inR _ True] k_suc show ?thesis by auto
         next
         case False
          with 1 have sk:"?k = Suc j'" by auto
          with xj x show "x \<in> ?xs!(Suc j')" unfolding X_part_last[OF inR nat.discI[OF sk], unfolded sk] by auto
        qed
        from fun_merge_part[OF p _ this] lx 1 have sxSucj:"?\<sigma> x = (\<sigma>!Suc j') x" by auto
       { fix j''
         assume 1:"Suc j' \<le> j''" and 2:"j'' < Suc j"
         with ik have ltk:"Suc j' < ?k" by force
         have "x \<in> set (X \<rho> j'')" proof(cases "Suc j' = j''", insert xsj[OF ltk],simp)
           case False with 1 have "Suc j' < j''" by auto
             from xsj[OF ltk]  X_vars_mono[OF inR _ 1] 2 ik show "x \<in> set (X \<rho> j'')" by auto
           qed 
         from X_Y_imp_Z[OF inR _ this y[OF 2[unfolded less_Suc_eq_le]]] 2 ik have "x \<in> set (Z \<rho> j'')" by auto
       } note z = this
       have "((\<sigma> ! Suc j') x,  (\<sigma> ! (Suc j)) x) \<in> (cstep R)\<^sup>*"
       proof(cases "Suc j' = Suc j", insert ji, simp)
       case False 
        with ji have "Suc j' < Suc j" by auto
        with z ik show ?thesis proof (induct j,simp)
        case (Suc k) 
         from Suc have z:"\<And> j''. Suc j' \<le> j'' \<and> j'' < Suc k \<Longrightarrow> x \<in> set (Z \<rho> j'')" by auto
         show ?case proof(cases "Suc j'= Suc k")
         case False
          with Suc(4) Suc(3) have 0:"j' < Suc k" and 1:"Suc k < ?k" and 2:"Suc j' < Suc k" by auto
          from z Suc(1)[OF _ less_imp_le_nat[OF 1] 2] have "((\<sigma> ! Suc j') x, (\<sigma> ! Suc k) x) \<in> (cstep R)\<^sup>*" by blast
          with zrew[rule_format, OF 1] Suc(2)[of "Suc k", OF Suc_leI[OF 0] lessI] show ?thesis by force
         next
         case True
           from Suc have 0:"Suc j' < ?k" by auto
           from True zrew[rule_format, OF 0 Suc(2)[OF order.refl Suc(4)]] show ?thesis by simp
         qed
       qed
       qed
       with sxSucj show ?thesis by auto
    qed
 } note auxiliary2 = this
(* case for s_i *)
 { fix i x
   assume ik:"i < ?k" 
   { fix x
   assume xs:"x \<in> vars_term (si \<rho> i)"
   have " (?\<sigma> x, (\<sigma>!i) x) \<in> (cstep R)^*" proof(cases i) 
    case 0
     from xs[unfolded 0] dctrs[rule_format, OF inR ik[unfolded 0]] have  "x \<in> vars_term ?l" by fast
     with l_vars_X_vars[OF inR] k_suc have "x \<in> ?xs!0"  unfolding X_part_def nth_Cons_0 by auto
     from fun_merge_part[OF p _ this, unfolded lx[symmetric] k_suc] have  "?\<sigma> x = (\<sigma>!0) x" by auto
     with 0 show ?thesis by auto
    next
    case (Suc j)
     from s_imp_Y[OF inR ik[unfolded Suc] xs[unfolded Suc]] ik have y:" x \<in> set (Y \<rho> j)" by blast
     from rev_subsetD[OF xs dctrs[rule_format, OF inR ik]] auxiliary2[OF y] Suc ik show ?thesis by fastforce
   qed
   } 
   with ik substs_rsteps' have "((si \<rho> i) \<cdot> ?\<sigma>, (si \<rho> i) \<cdot> (\<sigma>!i)) \<in> (cstep R)^*" by blast
 } note ss = this
(* case for r *)
 { fix i x
   assume xr:"x \<in> vars_term (crhs \<rho>)"
   from rev_subsetD[OF xr type3[rule_format, OF inR], unfolded Un_iff[of x]] have 
      disj:"x \<in> vars_term ?l \<or> (\<exists> j. j< ?k \<and> x \<in> vars_term (ti \<rho> j))" by fastforce
   from conjunct1[OF conjunct2[OF Z[unfolded Z_vars_def]], unfolded Y_vars_def, rule_format, OF inR] xr
     have y:" x \<in> set (Y \<rho> k')" unfolding k_suc by simp
   from auxiliary2[OF y] disj k_suc have "(?\<sigma> x, (\<sigma>!?k) x) \<in> (cstep R)^*"  by auto
} note xr = this
 from xr substs_rsteps' have "(?r \<cdot> ?\<sigma>, ?r \<cdot> (\<sigma>!?k)) \<in> (cstep R)^*" by blast 
 then obtain m where rsigma:"(?r \<cdot> ?\<sigma>, ?r \<cdot> (\<sigma>!?k)) \<in> (cstep_n R m)^*" using csteps_imp_csteps_n by blast
(* combine to final result *) 
 from ss st[rule_format] t_fact have "\<forall> i  < ?k.((si \<rho> i) \<cdot> ?\<sigma>,(ti \<rho> i) \<cdot> ?\<sigma>) \<in> (cstep R)^*" 
  using rtrancl_trans[of _ _ "cstep R"] nat_less_le by metis 
 from all_cstep_imp_cstep_n[of "snd \<rho>", OF this] obtain n where "\<forall> i < ?k. ((si \<rho> i) \<cdot> ?\<sigma>,(ti \<rho> i) \<cdot> ?\<sigma>) \<in> (cstep_n R n)^*" by blast
 hence "\<forall>(si, ti) \<in> set (snd \<rho>). (si \<cdot> ?\<sigma>, ti \<cdot> ?\<sigma>) \<in> (cstep_n R n)\<^sup>*"
   by (auto) (metis fst_conv in_set_idx snd_conv)
 from cstep_n_SucI [OF rho, of _ _ "?l \<cdot> ?\<sigma>" Hole "rhs_n \<rho> ?k \<cdot> ?\<sigma>", unfolded ctxt_apply.simps(1), OF this, unfolded lsigma] rsigma rhs_n.simps k_suc
   have "(?l \<cdot> (\<sigma> ! 0), rhs_n \<rho> ?k \<cdot> ?\<sigma>) \<in> (cstep_n R (Suc n))\<^sup>+" by (metis less_not_refl prod.collapse r_into_trancl') 
 from this[unfolded rhs_n.simps[of "fst \<rho>" "snd \<rho>", unfolded prod.collapse] k_suc] have  "(?l \<cdot> (\<sigma> ! 0), ?r \<cdot> ?\<sigma>) \<in> (cstep_n R (Suc n))\<^sup>+" by force
 from csteps_n_trans[OF this rsigma] obtain n where q:"(clhs \<rho> \<cdot> \<sigma> ! 0, crhs \<rho> \<cdot> \<sigma> ! length (snd \<rho>)) \<in> (cstep_n R n)\<^sup>+" by auto 
 from trancl_mono_set[of "cstep_n R n" "\<Union>n. cstep_n R n"] q show ?thesis unfolding cstep_def by blast
qed

definition non_LV :: bool where "non_LV \<equiv> \<forall> (l,r) \<in> UR. \<not>is_Var l"

definition sig_R :: "('f,'v) term \<Rightarrow> bool" where "sig_R t \<equiv> (funs_term t \<subseteq> funs_ctrs R)"

abbreviation sig_R_subst :: "('f,'v) subst \<Rightarrow> 'v set \<Rightarrow> bool" where 
 "sig_R_subst \<theta> V \<equiv> (\<forall> x \<in> V. sig_R (\<theta> x))"

lemma ur_step_exhaust:
 assumes step:"(Fun f us,v) \<in> rstep_r_p_s UR (l,r) \<epsilon> \<sigma>" and nlv:"non_LV"
 shows "(\<exists> \<rho> \<in> R. ((l, r) = (lhs_n \<rho> 0, rhs_n \<rho> 0) \<and> f \<in> funs_ctrs R) \<or> 
        (\<exists> n < length (snd \<rho>).((l, r) = (lhs_n \<rho> (Suc n), rhs_n \<rho> (Suc n)) \<and> U \<rho> n = More f [] \<box> (map Var (Z \<rho> n)) \<and> f \<notin> funs_ctrs R)))"
proof-
 from step[unfolded rstep_r_p_s_def] have ur:"(l,r) \<in> UR" by auto
 from this[unfolded UR_def] rules_def obtain n \<rho> where n:"\<rho> \<in> R" "(l,r) = (lhs_n \<rho> n, rhs_n \<rho> n)""n \<le> length (snd \<rho>)" by blast 
 with PairE obtain s t cs where rho:"\<rho> = ((s,t), cs)" by metis
 from nlv[unfolded non_LV_def, rule_format, OF ur, unfolded split] obtain g ls where l:"l = Fun g ls" by blast
 with step[unfolded rstep_r_p_s_def ctxt_of_pos_term.simps(1) Let_def] have gf:"g = f" by simp
 show ?thesis proof (cases n)
  case 0
   from n(2)[unfolded 0 rho lhs_n.simps(1)] have ls:"s=l" by simp 
   from n(1)[unfolded rho ls l[unfolded gf]] have "f \<in> funs_ctrs R" unfolding funs_ctrs_def funs_rule_def by force
   with n 0 show ?thesis by fast 
  next
  case (Suc m) 
   with lhs_n.elims[OF conjunct1[OF n(2)[unfolded Pair_eq], symmetric]] rho have l':"l = (U \<rho> m)\<langle>snd (cs ! m)\<rangle>" by blast
   from Suc n(3) have m:"m < length (snd \<rho>)" by fastforce
   from U_cond[rule_format, OF n(1) this] obtain h where u:"U \<rho> m = More h [] \<box> (map Var (Z \<rho> m)) \<and> h \<notin> funs_ctrs R" by auto
   with l[unfolded gf] l' have "h=f" by simp
   from n(1) n(2)[unfolded Suc] u[unfolded this] m show ?thesis by auto
 qed
qed

lemma ur_step_urule:
 assumes step:"(Fun f us,v) \<in> rstep_r_p_s UR (l,r) \<epsilon> \<sigma>" 
 and nlv:"non_LV" 
 and rho:"\<rho> \<in> R" and n:"n < length (snd \<rho>)" 
 and u:"U \<rho> n = More f [] \<box> (map Var (Z \<rho> n))"
 shows "(l, r) = (lhs_n \<rho> (Suc n), rhs_n \<rho> (Suc n)) \<and> f \<notin> funs_ctrs R"
proof-
 from U_cond[rule_format, OF rho n] u have "f \<notin> funs_ctrs R" by force
 with ur_step_exhaust[OF step nlv] obtain \<rho>' m where d:"\<rho>' \<in> R" "m < length (snd \<rho>')"
  "(l, r) = (lhs_n \<rho>' (Suc m), rhs_n \<rho>' (Suc m))" "U \<rho>' m = More f [] \<box> (map Var (Z \<rho>' m))" "f \<notin> funs_ctrs R" by blast
 from U_cond[rule_format, OF d(1) d(2)] rho n u d(4) have "m = n \<and> \<rho> = \<rho>'" by fastforce
 with d show ?thesis by force
qed

lemma Um_funs_R:
 assumes um:"U \<rho> m = (More Um Nil Hole (map Var (Z \<rho> m)))" and len:"m < length (snd \<rho>)"
 and rhoR:"\<rho> \<in> R"
 shows "Um \<notin> funs_ctrs R"
 using U_cond[rule_format, OF rhoR len] um by auto
   

lemma Um_unique:
 assumes um:"U \<rho> m = (More Um Nil Hole (map Var (Z \<rho> m)))" and lm:"m < length (snd \<rho>)"
 assumes uk:"U \<rho> k = (More Um Nil Hole (map Var (Z \<rho> k)))" and lk:"k < length (snd \<rho>)"
 and rhoR:"\<rho> \<in> R"
 shows "m = k"
proof-
 from U_cond[rule_format, OF rhoR lm] um have " \<not> (\<exists>\<rho>' n' b c a. \<rho>' \<in> R \<and> n' < length (snd \<rho>') \<and> (\<rho> \<noteq> \<rho>' \<or> m \<noteq> n') \<and> U \<rho>' n' = More Um b c a)" by auto
 from this[unfolded not_ex, rule_format, of \<rho> k "[]" \<box> "map Var (Z \<rho> k)"] rhoR uk lk show ?thesis by fast
qed

lemma no_vars_in_seq:
 assumes seq:"\<forall>i<length us - 1. (us ! i, us ! Suc i) \<in> par_rstep UR"
 and nonlv:non_LV
 and l:"l < length us"
 and last:"us ! l = Fun f vs"
 and i:"i < l"
 shows "is_Fun (us ! i)"
proof(rule ccontr)
 assume "\<not> is_Fun(us ! i)"
 then obtain x where ix:"us ! i = Var x" by blast
 from nonlv[unfolded non_LV_def] have aux:"\<And> s r \<sigma>. (s,r) \<in> UR \<Longrightarrow> \<not> is_Var (s \<cdot> \<sigma>)" by fastforce
 { fix j 
   assume ji:"j \<ge> i" and jl:"j \<le> l"
   then have "us ! j = Var x" proof(induct rule:dec_induct[of i j])
    assume "i \<le> l"
    with ix show "us ! i = Var x" by auto
   next
    fix n
    assume Suc:"i \<le> n" "n < j" "n \<le> l \<Longrightarrow> us ! n = Var x" "Suc n \<le> l"
    hence n:"n < l" and n':"n \<le> l" by auto
    with Suc(3) have "us ! n = Var x" by force
    from n l have "n < length us - 1" by auto
    from seq[rule_format, OF this, unfolded Suc(3)[OF n']] par_rstep.simps[of "Var x"] aux is_VarI[of x]
     term.distinct(1)[of x] show "us ! (Suc n) = Var x" by metis
   qed 
 } from this[OF _ le_refl] i last l show False by auto
qed



lemma no_vars_in_seq':
 assumes seq:"\<forall>i<k - 1. (ui i, ui (Suc i)) \<in> par_rstep UR"
 and nonlv:non_LV
 and last:"ui (k - 1) = Fun f vs"
 and i:"i < k - 1" and k:"k > 0"
 shows "is_Fun (ui i)" 
proof-
 let ?us = "map ui [0..<k]"
 have l:"length ?us = k" by simp
 from seq nth_map[of _ "[0..<k]" ui] have 1:"\<forall>i<length ?us - 1. (?us ! i, ?us ! Suc i) \<in> par_rstep UR" unfolding l by force
 from last  nth_map_upt[of "k - 1" k 0 ui] k
  have 2:"?us ! (length ?us - 1) = Fun f vs" unfolding l by force
 from no_vars_in_seq[OF 1 nonlv _ 2, unfolded l] nth_map_upt[of i k 0 ui] i show ?thesis by force
qed

definition par_urstep_below_root :: "(('f,'v) term \<times> ('f,'v) term) set" where
 "par_urstep_below_root  \<equiv> {(s,t). \<exists>ts ss f. s = Fun f ss \<and> t = Fun f ts \<and> 
     (\<forall>x<length ts. (ss ! x, ts ! x) \<in> par_rstep UR) \<and> length ss = length ts}"

lemma par_urstep_par_rstep:
 "par_urstep_below_root \<subseteq> par_rstep UR" 
 unfolding par_urstep_below_root_def by auto


(* move *)
lemma all_ctxt_closed_rel_pow:
 assumes acc:"all_ctxt_closed UNIV Q"
 shows "all_ctxt_closed UNIV (Q ^^ n)" 
proof-
  { fix f ss ts
    assume  a: "length ts = length ss" "\<forall>i<length ts. (ts ! i, ss ! i) \<in> Q ^^ n"
    hence "(Fun f ts, Fun f ss) \<in> Q ^^ n" proof(induct n arbitrary: ss ts)
     case 0
      with relpow_0_I list_eq_iff_nth_eq relpow_0_E show ?case by metis
     next 
     case (Suc j)
       { fix i
         assume "i < length ts"
         from relpow_Suc_D2[OF Suc(3)[rule_format, OF this]] have "\<exists> u. (ts ! i, u) \<in> Q \<and> (u, ss ! i) \<in> Q ^^ j" by auto
       }
      with Suc(3) have "\<forall> i. \<exists> u. i < length ts \<longrightarrow> (ts ! i, u) \<in> Q \<and> (u, ss ! i) \<in> Q ^^ j" by blast
      from choice[OF this] obtain ui where x:"\<forall>i < (length ts). (ts ! i, ui i) \<in> Q \<and> (ui i, ss ! i) \<in> Q ^^ j" by blast
      let ?us = "map ui [0..<length ts]"
      from x have 1:"\<forall>i < (length ts). (ts ! i, ?us ! i) \<in> Q \<and> (?us ! i, ss ! i) \<in> Q ^^ j" by auto
      from Suc(2) have 2:"length ?us = length ss" by simp
      with Suc(2) have 3:"length ts = length ?us" by auto
      from Suc(1)[OF 2] 1 have y:"(Fun f ?us, Fun f ss) \<in> Q ^^ j" by fastforce
      from conjunct1[OF acc[unfolded all_ctxt_closed_def], rule_format, OF _ 3] x have "(Fun f ts, Fun f ?us) \<in> Q" by fastforce
      from relpow_Suc_I2[OF this y] show ?case by blast
    qed
  } note 1 = this
  from assms[unfolded all_ctxt_closed_def] have "\<forall>x. (Var x, Var x) \<in> Q ^^ n" by (induct n, auto)
  with 1 show ?thesis unfolding all_ctxt_closed_def by simp
qed

lemma par_urstep_below_root_refl: "(Fun f us, Fun f us) \<in> par_urstep_below_root" 
 unfolding par_urstep_below_root_def by blast

  
lemma funs_term_subterm_subset:
 shows "t  \<unrhd> s \<Longrightarrow> funs_term s \<subseteq> funs_term t" 
 by (induct rule:supteq.induct, auto)

(* move? *)
lemma relpow_refl_mono:
 assumes refl:"\<forall> x. (x,x) \<in> Rel"
 shows " m \<le> n \<Longrightarrow>(a,b) \<in> Rel ^^ m \<Longrightarrow> (a,b) \<in> Rel ^^ n"
proof (induct rule:dec_induct, simp)
 fix i
 assume "m \<le> i" and "i < n" and "(a, b) \<in> Rel ^^ m \<Longrightarrow> (a, b) \<in> Rel ^^ i" and "(a, b) \<in> Rel ^^ m"
 hence abi:"(a, b) \<in> Rel ^^ i" by auto
 from refl[rule_format, of b] abi relpowp_Suc_I[of i "\<lambda> x y. (x,y) \<in> Rel"] show "(a, b) \<in> Rel ^^ Suc i" by auto
qed

lemma args_par_rstep_pow_imp_par_rstep_pow:
 "length xs = length ys \<Longrightarrow> \<forall>i<length xs. (xs ! i, ys ! i) \<in> par_rstep Q ^^ n \<Longrightarrow>
  (Fun f xs, Fun f ys) \<in> par_rstep Q ^^ n"
proof(induct n arbitrary:ys)
 case 0 
  hence "\<forall>i<length xs. (xs ! i = ys ! i)" by simp
  with 0 show ?case using relpow_0_I list_eq_iff_nth_eq by metis
 next
 case (Suc n)
 let ?c = "\<lambda> z i. (xs ! i, z) \<in> par_rstep Q ^^ n \<and> (z, ys ! i) \<in> par_rstep Q"
  { fix i assume "i < length xs"
    from relpow_Suc_E[OF Suc(3)[rule_format, OF this]] 
     have "\<exists> z. (?c z i)" by metis
  }
  with choice have "\<exists> zf. \<forall> i < length xs. (?c (zf i) i)" by meson
  then obtain zf where a:"\<forall> i < length xs. (?c (zf i) i)" by auto
  let ?zs = "map zf [0..<length xs]"
  have len:"length xs = length ?zs" by simp
  from a map_nth have  "\<forall>i<length xs.  (xs ! i, ?zs ! i) \<in> par_rstep Q ^^ n" by auto
  from Suc(1)[OF len this] have n:"(Fun f xs, Fun f ?zs) \<in> par_rstep Q ^^ n" by auto
  from a map_nth have  "\<forall>i<length xs. (?zs ! i, ys ! i) \<in> par_rstep Q" by auto
  with par_step_fun len Suc(2) have "(Fun f ?zs, Fun f ys) \<in> par_rstep Q" by auto
  with n show ?case by auto
qed

lemma par_rstep_below_root_pow_same_root:
 "(Fun f xs, t) \<in> par_urstep_below_root ^^ n \<Longrightarrow> \<exists> ys. t = Fun f ys"
 unfolding par_urstep_below_root_def by (induct n arbitrary: t, auto)

lemma par_rstep_below_root_pow_same_length: (* currently unused *)
 "(Fun f xs, Fun f ys) \<in> par_urstep_below_root ^^ n \<Longrightarrow> length xs = length ys"
 unfolding par_urstep_below_root_def by (induct n arbitrary: ys, auto)

lemma par_rstep_pow_imp_args_par_rstep_pow: (* currently unused *)
 "(Fun f xs, Fun f ys) \<in> par_urstep_below_root ^^ n = ((\<forall>i<length xs. (xs ! i, ys ! i) \<in> par_rstep UR ^^ n) \<and> length xs = length ys)"
proof
 assume "(Fun f xs, Fun f ys) \<in> par_urstep_below_root ^^ n"
 thus "((\<forall>i<length xs. (xs ! i, ys ! i) \<in> par_rstep UR ^^ n) \<and> length xs = length ys)"
 proof(induct n arbitrary:ys)
 case 0
  then show ?case unfolding par_urstep_below_root_def by auto
 next
 case (Suc n)
  from  relpow_Suc_E[OF Suc(2)] par_rstep_below_root_pow_same_root obtain zs where 
   steps:"(Fun f xs, Fun f zs) \<in> par_urstep_below_root ^^ n" "(Fun f zs, Fun f ys) \<in> par_urstep_below_root" by metis
   from steps(1) par_rstep_below_root_pow_same_length have len:"length xs = length zs" by auto
   from CollectD[OF steps(2)[unfolded par_urstep_below_root_def], unfolded split] have len':"length ys = length zs" by fastforce
  {fix i
   assume i:"i<length xs" 
   from Suc(1) steps(1) this have xs:"(xs ! i, zs ! i) \<in> par_rstep UR ^^ n" by auto
   from CollectD[OF steps(2)[unfolded par_urstep_below_root_def], unfolded split] i[unfolded len]
       have "(zs ! i, ys ! i)  \<in> par_rstep UR" by fastforce
   with xs have "(xs ! i, ys ! i) \<in> par_rstep UR ^^ (Suc n)" by auto
  } with len len' show ?case by auto
 qed
 next
  assume "(\<forall>i<length xs. (xs ! i, ys ! i) \<in> par_rstep UR ^^ n) \<and> length xs = length ys"
  thus "(Fun f xs, Fun f ys) \<in> par_urstep_below_root ^^ n"
  proof(induct n arbitrary:ys)
   case 0 
    hence "\<forall>i<length xs. xs ! i = ys ! i" by simp 
    with 0 list_eq_iff_nth_eq have "xs = ys" by blast
    with 0 show ?case unfolding par_urstep_below_root_def by simp
   next
   case (Suc n)
    {fix i
     assume i:"i<length xs" 
     from conjunct1[OF Suc(2), rule_format, OF i] relpow_Suc_E have
      "\<exists>z. (xs ! i, z) \<in> par_rstep UR ^^ n \<and> (z, ys ! i) \<in> par_rstep UR" by auto
    }
    then obtain fz where fz:"(\<forall>i < length xs. (xs ! i, fz i) \<in> par_rstep UR ^^ n \<and> (fz i, ys ! i) \<in> par_rstep UR )" by metis
    let ?zs = "map fz [0..<length xs]"
    from Suc have len:"length ?zs = length ys" by simp
    from fz have zs:"(\<forall>i < length xs. (xs ! i, ?zs ! i) \<in> par_rstep UR ^^ n \<and> (?zs ! i, ys ! i) \<in> par_rstep UR)" by force
    with Suc have n:"(Fun f xs, Fun f ?zs) \<in> par_urstep_below_root ^^ n" by fastforce
    from zs len have "(Fun f ?zs, Fun f ys) \<in> par_urstep_below_root" unfolding par_urstep_below_root_def by force
    with n show ?case by auto
   qed
qed

lemma ur_sequence2:
 assumes seq:"(\<forall>i<(length us) - 1. (us ! i, us ! Suc i) \<in> par_rstep UR)"
 and no_last_rule:"\<forall>i. (start \<le> i \<and> i < fin) \<longrightarrow> \<not> (\<exists> \<tau> . 
  (us ! i, us ! Suc i) \<in> rstep_r_p_s UR (lhs_n \<rho> (length (snd \<rho>)), rhs_n \<rho> (length (snd \<rho>))) \<epsilon> \<tau>)"
 and rhoR:"\<rho> \<in> R"
 and uroot:"us ! start = Fun Um ws" "U \<rho> m = (More Um Nil Hole (map Var (Z \<rho> m)))" "m < length (snd \<rho>)"
 and sl:"start \<le> fin \<and> fin < length us"
 and nonlv:"non_LV"
 and len:"length us > 0"
 shows "(us ! start, us ! fin) \<in> par_urstep_below_root ^^ (fin - start) \<or>
 (\<exists> j ni nm nk k \<sigma>i. 
 m < k \<and> k <  length (snd \<rho>) \<and> (\<exists> ws' Uk. us ! fin = Fun Uk ws' \<and> U \<rho> k = (More Uk Nil Hole (map Var (Z \<rho> k)))) \<and> 
 (\<forall>i. m < i \<and> i \<le> k \<longrightarrow> start \<le> j i \<and> j i < fin \<and>
   (us ! (j i), us ! Suc (j i)) \<in> (rstep_r_p_s UR (lhs_n \<rho> i, rhs_n \<rho> i) \<epsilon> (\<sigma>i i))) \<and> 
 (us ! start, us ! (j (Suc m))) \<in>  par_urstep_below_root ^^ nm \<and>
 (\<forall>i. m < i \<and> i < k \<longrightarrow> (us ! Suc (j i), us ! (j (Suc i))) \<in> par_urstep_below_root ^^ (ni i)) \<and>
 ( (us ! Suc (j k), us ! fin) \<in> par_urstep_below_root ^^ nk) \<and> 
 start + nm + (listsum (map ni [Suc m..<k])) + (k - m) + nk = fin)" 
proof-
 from seq no_last_rule uroot sl len show ?thesis proof(induct "length us" arbitrary: us Um ws m start fin)
  case 0 
   thus ?case by force
  next
  case (Suc n)
   let ?no_last = "\<lambda> s us f. \<forall>i. (s \<le> i \<and> i < f) \<longrightarrow> \<not> (\<exists> \<tau> . 
     (us ! i, us ! Suc i) \<in> rstep_r_p_s UR (lhs_n \<rho> (length (snd \<rho>)), rhs_n \<rho> (length (snd \<rho>))) \<epsilon> \<tau>)"
   let ?same' = "\<lambda> s f us ws ws' Um.  us ! f = Fun Um ws' \<and> length ws = length ws' \<and> (\<forall>j < length ws. (ws ! j, ws' ! j) \<in> par_rstep UR ^^ (f - s))"
   let ?same = "\<lambda> s f us. (us ! s, us ! f) \<in> par_urstep_below_root ^^ (f - s) "
   let ?ex_k' = "\<lambda> m k us f. m < k \<and> k < length (snd \<rho>) \<and> (\<exists> ws' Uk. us ! f = Fun Uk ws' \<and> U \<rho> k = (More Uk Nil Hole (map Var (Z \<rho> k))))" (* 1 *)
   let ?step = "\<lambda> ts ji i \<sigma> . (ts ! ji, ts ! Suc ji) \<in> (rstep_r_p_s UR (lhs_n \<rho> i, rhs_n \<rho> i) \<epsilon> \<sigma>)"
   let ?ur_step = "?step us" and ?ur_step' = "?step (tl us)"
   let ?ur_steps' = "\<lambda> m k s us j f \<sigma>i. \<forall>i. (m < i \<and> i \<le> k \<longrightarrow> s \<le> j i \<and> j i < f \<and> ?step us (j i) i (\<sigma>i i))"                              (* 2 *)
   let ?pstep0' = "\<lambda> m s us j nm. (us ! s, us ! (j (Suc m))) \<in>  par_urstep_below_root ^^ nm"                                               (* 3 *)
   let ?psteps' = "\<lambda> m k us j ni. (\<forall>i. m < i \<and> i < k \<longrightarrow> (us ! Suc (j i), us ! (j (Suc i))) \<in> par_urstep_below_root ^^ (ni i))"            (* 4 *)
   let ?pstepk = "\<lambda> k us j nk f. (us ! Suc (j k), us ! f) \<in> par_urstep_below_root ^^ nk"                                               (* 5 *)
   let ?sum' = "\<lambda> m k st us nm ni nk f. st + nm + (listsum (map ni [Suc m..<k])) + (k - m) + nk = f"                                         (* 6 *)
   let ?aim' = "\<lambda> m k st us j nm ni nk f \<sigma>i. (?ex_k' m k us f \<and> ?ur_steps' m k st us j f \<sigma>i \<and> ?pstep0' m st us j nm \<and> ?psteps' m k us j ni \<and> 
               ?pstepk k us j nk f \<and> ?sum' m k st us nm ni nk f)"
   let ?ex_k = "?ex_k' m" and ?ur_steps = "?ur_steps' m" and ?pstep0 = "?pstep0' m" and ?psteps = "?psteps' m" and ?sum = "?sum' m" and ?aim = "?aim' m"
   show ?case proof (cases "start = fin")
    case True
     from Suc(5)[unfolded this] have "?same start fin us" unfolding True by auto
     with True show ?thesis by auto
    next
    case False 
     with Suc(8) have sf:"start < fin" by auto
     let ?us = "tl us"
     from False Suc(2) have 1: "length us  = Suc n" "length ?us = n" by auto
     from sf Suc(8) have n0:"0 < n" unfolding 1 by auto
     from gr0_conv_Suc sf Suc(8) obtain fin' where fin:"fin = Suc fin'" and fn:"fin' < n" unfolding 1 by auto
     from Um_funs_R[OF Suc(6) Suc(7) rhoR] have UmR:"Um \<notin> funs_ctrs R" by auto
     note nth_tl' = nth_tl[of _ us, unfolded 1]
     from Suc(3) Suc(2) nth_tl' have 2:"(\<forall>i < n - 1. (?us ! i, ?us ! Suc i) \<in> par_rstep UR)" by fastforce
     show ?thesis proof(cases "start > 0")
     case True 
      with gr0_conv_Suc Suc(8) Suc(2) obtain start' where start:"start = Suc start'" and sn:"start' < n" by auto 
      from Suc(8) start fin have sf:"start' \<le> fin'" by auto
      from Suc(4) Suc(8) nth_tl' 1 have 3:"?no_last start' ?us fin'" unfolding start fin by auto
      from sn have n0:"0 < n" by linarith
      from nth_tl'[OF sn] Suc(5) have 4:"?us ! start' = Fun Um ws" unfolding start by auto
      from Suc(1)[OF 1(2)[symmetric], unfolded 1, OF 2 3 4 Suc(6) Suc(7) _ n0] sf fn have 
         alts:"(\<exists> ws'. ?same start' fin' ?us) \<or> (\<exists> j ni nm nk k \<sigma>i. ?aim k start' ?us j nm ni nk fin' \<sigma>i)" by auto
      { assume "\<exists> ws'. ?same start' fin' ?us"
        with nth_tl'[OF fn] nth_tl'[OF sn] have "\<exists> ws'. ?same start fin us" unfolding fin start by auto
      } note same = this
      { assume "\<exists> j ni nm nk k \<sigma>i. ?aim k start' ?us j nm ni nk fin' \<sigma>i"
        then  obtain j' ni nm nk k \<sigma>i where 
        ih:"?ex_k k ?us fin'""?ur_steps k start' ?us j' fin' \<sigma>i" "?pstep0 start' ?us j' nm" "?psteps k ?us j' ni" 
          "?pstepk k ?us j' nk fin'" "?sum k start' ?us nm ni nk fin'" unfolding 1 by blast
        let ?j = "\<lambda> i. Suc (j' i)"
        from ih(2)[rule_format] fn have aux:"\<And>i. m < i \<and> i \<le> k \<Longrightarrow> Suc (j' i) < n" by force
        from ih(2)[rule_format] fn have aux':"\<And>i. m < i \<and> i \<le> k \<Longrightarrow> j' i < n" by force
        from ih(1) nth_tl'[OF fn] have 1:"?ex_k k us fin" unfolding fin by auto
        { fix i
          assume i:"m < i \<and> i \<le> k"
          from ih(2)[rule_format, OF i] fn have "j' i < n""Suc (j' i) < n" by auto
          from ih(2)[rule_format, OF i] fn nth_tl'[OF this(1)]nth_tl'[OF this(2)] have 
           "start \<le> ?j i \<and> ?j i < fin \<and> ?step us (?j i) i (\<sigma>i i)" unfolding start fin by auto
        } note 2 = this
        from ih(1) ih(2)[rule_format, of "Suc m"] fn have "j' (Suc m) < n" by auto
        from ih(3) nth_tl'[OF sn] nth_tl'[OF this] have 3:"?pstep0 start us ?j nm" unfolding start by auto
        from ih(4) nth_tl'[OF aux] nth_tl'[OF aux'] have 4:"?psteps k us ?j ni" by simp
        from ih(1) ih(5) nth_tl'[OF aux, of k] nth_tl'[OF fn] have 5:"?pstepk k us ?j nk fin" unfolding fin by force
        from ih(6) have 6:"?sum k start us nm ni nk fin" unfolding start fin by auto
        have "\<exists> j ni nm nk k \<sigma>i. ?aim k start us j nm ni nk fin \<sigma>i" by(rule exI[of _ ?j],rule exI[of _ ni],  insert 1 2 3 4 5 6, blast)
       } with same alts show ?thesis by presburger
     next
     case False
      hence s0:"start = 0" by auto
      show ?thesis proof(cases "\<exists> vs. us ! (Suc 0) = Fun Um vs \<and> (\<forall>i < length vs. (ws ! i, vs ! i) \<in> par_rstep UR) \<and> length ws = length vs")
      case True (* case 1:  first step preserves root symbol *)
       then obtain vs where t:"us ! (Suc 0) = Fun Um vs" "\<forall>i < length vs. (ws ! i, vs ! i) \<in> par_rstep UR" "length ws = length vs" by auto
       from t(1) Suc(8) nth_tl[of 0 us, unfolded 1] sf have 4:"?us ! 0 = Fun Um vs" unfolding 1 s0 by force
       from Suc(4) Suc(8) nth_tl' 1 have 3:"?no_last 0 ?us fin'" unfolding s0 fin by simp
       from Suc(1)[OF 1(2)[symmetric], unfolded 1, OF 2 3 4 Suc(6) Suc(7) _ n0] fn have 
         alts:"?same 0 fin' ?us \<or> (\<exists> j ni nm nk k \<sigma>i. ?aim k 0 ?us j nm ni nk fin' \<sigma>i)" by auto
       { assume a:"?same 0 fin' ?us" 
         with 4 par_rstep_below_root_pow_same_root obtain ws' where last:"?us ! fin' = Fun Um ws'" by force
         from a 4 this par_rstep_pow_imp_args_par_rstep_pow have same: "?same' 0 fin' ?us vs ws' Um" unfolding s0 by force
         { fix j assume "j < length ws"
           with same  relpow_Suc_I2[OF t(2)[unfolded  t(3)[symmetric], rule_format, OF this]] 
             have "(ws ! j, ws' ! j) \<in> par_rstep UR ^^ fin" unfolding fin t(3)[symmetric] by force
         }
         with same t(3) have "?same' 0 fin us ws ws' Um" unfolding fin nth_tl'[OF fn] by simp
         with Suc(5)[unfolded s0] last[unfolded nth_tl'[OF fn]] have "?same 0 fin us" using par_rstep_pow_imp_args_par_rstep_pow unfolding fin by presburger
       } note same = this
       { assume "\<exists> j ni nm nk k \<sigma>i. ?aim k 0 ?us j nm ni nk fin' \<sigma>i"
         then obtain j' ni nm' nk k \<sigma>i where 
         ih:"?ex_k k ?us fin'""?ur_steps k 0 ?us j' fin' \<sigma>i" "?pstep0 0 ?us j' nm'" "?psteps k ?us j' ni" 
          "?pstepk k ?us j' nk fin'" "?sum k 0 ?us nm' ni nk fin'" unfolding 1 by blast
        let ?j = "\<lambda> i. Suc (j' i)" and ?nm = "Suc nm'"
        from ih(2)[rule_format] fn have aux:"\<And>i. m < i \<and> i \<le> k \<Longrightarrow> Suc (j' i) < n" by force
        from ih(2)[rule_format] fn have aux':"\<And>i. m < i \<and> i \<le> k \<Longrightarrow> j' i < n" by force
        from ih(1) nth_tl'[OF fn] have 2:"?ex_k k us fin" unfolding fin by auto
        { fix i
          assume i:"m < i \<and> i \<le> k"
          from ih(2)[rule_format, OF i] fn have "j' i < n""Suc (j' i) < n" by auto
          from ih(2)[rule_format, OF i] fn nth_tl'[OF this(1)]nth_tl'[OF this(2)] have 
           "start \<le> ?j i \<and> ?j i < fin \<and> ?step us (?j i) i (\<sigma>i i)" unfolding s0 fin by auto
        } note 3 = this
        from ih(1) ih(2)[rule_format, of "Suc m"] fn have jn:"j' (Suc m) < n" by auto 
        from Suc(5) t have "(us ! 0, us ! Suc 0) \<in> par_urstep_below_root" unfolding s0 par_urstep_below_root_def by auto
        with ih(3) n0 relpow_Suc_I2[OF this] have 4:"?pstep0 start us ?j ?nm" unfolding nth_tl'[OF n0] nth_tl'[OF jn] s0 by blast
        from ih(4) nth_tl'[OF aux] nth_tl'[OF aux'] have 5:"?psteps k us ?j ni" by simp
        from ih(1) ih(5) nth_tl'[OF aux, of k] nth_tl'[OF fn] have 6:"?pstepk k us ?j nk fin" unfolding fin by force
        from ih(6) have 7:"?sum k 0 us ?nm ni nk fin" unfolding s0 fin by auto
        from 1 2 3 4 5 6 7 have aim:"?aim k 0 us ?j ?nm ni nk fin \<sigma>i" unfolding s0 by blast
        have "\<exists> j ni nm nk k \<sigma>i. ?aim k 0 us j nm ni nk fin \<sigma>i" by (rule exI[of _ ?j],rule exI[of _ ni], insert aim, blast)
     } with same alts show ?thesis unfolding s0 by presburger
    next (* case 2: root symbol changes *)
    case False
     from Suc(3)[rule_format, of 0] n0 have step:"(us ! 0, us ! Suc 0) \<in> par_rstep UR" unfolding 1 by force
     from this[unfolded par_rstep.simps[of "us ! 0"], unfolded Suc(5)] False 1 n0 Suc(5)
      obtain lm rm \<sigma> where "Fun Um ws = lm \<cdot> \<sigma> \<and> us ! Suc 0 = rm \<cdot> \<sigma> \<and> (lm, rm) \<in> UR" unfolding s0 by auto
     with Suc(5) have step:"(us ! 0, us ! (Suc 0)) \<in> rstep_r_p_s UR (lm,rm) \<epsilon> \<sigma>" unfolding s0 rstep_r_p_s_def by force
     from ur_step_urule[OF step[unfolded Suc(5)[unfolded s0]] nonlv rhoR Suc(7) Suc(6)] have 
      lr:"(lm, rm) = (lhs_n \<rho> (Suc m), rhs_n \<rho> (Suc m))" "Um \<notin> funs_ctrs R" by auto
     with step have step:"?ur_step 0 (Suc m) \<sigma>" by auto
     hence rhs:"us ! 1 = rhs_n \<rho> (Suc m) \<cdot> \<sigma>" unfolding rstep_r_p_s_def by auto
     from PairE obtain l r cs where rho:"\<rho> = ((l,r), cs)" by metis
     from step[unfolded lr] Suc(4)[rule_format, unfolded s0, of 0] sf[unfolded s0] have "Suc m \<noteq> length (snd \<rho>)" by fastforce
     with Suc(7) have mk:"Suc m < length (snd \<rho>)" by auto
     with rhs rhs_n.simps[of "fst \<rho>" "snd \<rho>" "Suc m"] have us1:"us ! 1 = (U (fst \<rho>, snd \<rho>) (Suc m))\<langle>si \<rho> (Suc m)\<rangle> \<cdot> \<sigma>" by force
     from U_cond[rule_format, OF rhoR mk] obtain Um' where 
        x:"U \<rho> (Suc m) = More Um' [] \<box> (map Var (Z \<rho> (Suc m)))" "Um' \<notin> funs_ctrs R" by auto
     with us1[unfolded pair_collapse] obtain zs where Um':"us ! 1 = Fun Um' zs" by auto
     from nth_tl' n0 Um' have 5:"?us ! 0 = Fun Um' zs" by simp
     from Suc(4) Suc(8) nth_tl' 1 have 3:"?no_last 0 ?us fin'" unfolding s0 fin by simp
     from Suc(1)[OF 1(2)[symmetric], unfolded 1, OF 2 3 5 x(1) mk _ n0] fn have 
         alts:"(?same 0 fin' ?us) \<or> (\<exists> j ni nm nk k \<sigma>i. ?aim' (Suc m) k 0 ?us j nm ni nk fin' \<sigma>i)" by auto
     { assume a:"?same 0 fin' ?us"
       with 5 par_rstep_below_root_pow_same_root obtain ws' where last:"?us ! fin' = Fun Um' ws'" by force
       with x(1) nth_tl'[OF fn] mk have 2:"?ex_k (Suc m) us fin" unfolding fin by force
       from sf[unfolded s0] step have 3:"?ur_steps (Suc m) 0 us (\<lambda> i. 0) fin (\<lambda>i. \<sigma>)" by (metis le0 le_neq_implies_less not_less_eq)
       have 4:"?pstep0 0 us (\<lambda> i. 0) 0" by auto
       have 5:"?psteps (Suc m) us (\<lambda>i. 0) (\<lambda>i. 0)" by auto 
       from a nth_tl'[OF fn] nth_tl'[OF n0] have 6:"?pstepk (Suc m) us (\<lambda>i. 0) fin' fin" unfolding fin by force
       have 7:"?sum (Suc m) 0 us 0 (\<lambda>i. 0) fin' fin" unfolding fin by auto
       with 1 2 3 4 5 6 have aim:"?aim' m (Suc m) 0 us (\<lambda>i. 0) 0 (\<lambda>i.0) fin' fin (\<lambda>i. \<sigma>)" by blast
       have "\<exists> j ni nm nk k \<sigma>i. ?aim' m k 0 us j nm ni nk fin \<sigma>i" 
        by(rule exI[of _ "\<lambda>j. 0"], rule exI[of _ "\<lambda>j. 0"],rule exI[of _ "0"],rule exI[of _ fin'],rule exI[of _ "Suc m"], rule exI[of _ "\<lambda>j. \<sigma>"], insert aim, blast)
     } note same = this
     { assume "\<exists> j ni nm nk k \<sigma>i. ?aim' (Suc m) k 0 ?us j nm ni nk fin' \<sigma>i"
       then obtain j' ni nm' nk k \<sigma>i where 
        ih:"?ex_k' (Suc m) k ?us fin'""?ur_steps' (Suc m) k 0 ?us j' fin' \<sigma>i" "?pstep0' (Suc m) 0 ?us j' nm'" "?psteps' (Suc m) k ?us j' ni" 
          "?pstepk k ?us j' nk fin'" "?sum' (Suc m) k 0 ?us nm' ni nk fin'" unfolding 1 by blast
        let ?j = "\<lambda> i. if i = Suc m then 0 else Suc (j' i)" and ?nm = 0 and 
         ?ni = "\<lambda> i. if i = Suc m then nm' else ni i" and ?\<sigma>i = "\<lambda> i. if i = Suc m then \<sigma> else \<sigma>i i"
        from ih(2)[rule_format] fn have aux:"\<And>i. Suc m < i \<and> i \<le> k \<Longrightarrow> Suc (j' i) < n" by force
        from ih(2)[rule_format] fn have aux':"\<And>i. Suc m < i \<and> i \<le> k \<Longrightarrow> j' i < n" by force
        from ih(1) nth_tl'[OF fn] have 2:"?ex_k k us fin" unfolding fin by auto
        { fix i
          assume i:"m < i \<and> i \<le> k"
          have "start \<le> ?j i \<and> ?j i < fin \<and> ?step us (?j i) i (?\<sigma>i i)" proof (cases "Suc m < i")
          case True
           from ih(2)[rule_format, of i] True i fn have "j' i < n""Suc (j' i) < n" by auto
           from ih(2)[rule_format, of i] True i fn nth_tl'[OF this(1)]nth_tl'[OF this(2)] show ?thesis  unfolding s0 fin by auto
          next
          case False
           with i have i:"i = Suc m" by auto
           from sf step show ?thesis unfolding s0 i by auto
          qed
        } note 3 = this 
        have 4:"?pstep0 start us ?j ?nm" unfolding s0 simp_thms if_True by auto
        { fix i
          assume i:"m < i \<and> i < k"
          have "(us ! Suc (?j i), us ! ?j (Suc i)) \<in> par_urstep_below_root ^^ ?ni i" proof (cases "Suc m < i")
          case True
           with i ih(4)[rule_format, of i] nth_tl'[OF aux] nth_tl'[OF aux'] show ?thesis by auto
          next
          case False
           with i have i:"i = Suc m" by auto
           from ih(3) nth_tl'[OF n0] nth_tl'[OF aux', of "Suc (Suc m)"] ih(1) have 
            "(us ! (Suc 0), us ! Suc (j' (Suc (Suc m)))) \<in> par_urstep_below_root ^^ nm'" by auto
           then show ?thesis unfolding i by auto
          qed
        } note 5 = this
        from ih(1) ih(5) nth_tl'[OF aux, of k] nth_tl'[OF fn] have 6:"?pstepk k us ?j nk fin" unfolding fin by force
        from conjunct1[OF ih(1)] upt_conv_Cons list.map(2) have "map ?ni [Suc m..<k] = nm' # map ni [Suc (Suc m)..<k]" by simp
        hence s:"listsum (map ?ni [Suc m..<k]) = nm' + listsum (map ni [Suc (Suc m)..<k])" by auto
        from conjunct1[OF ih(1)] ih(6) have 7:"?sum k 0 us ?nm ?ni nk fin" unfolding s0 fin s by linarith
        from 1 2 3 4 5 6 7 have aim:"?aim k 0 us ?j 0 ?ni nk fin ?\<sigma>i" unfolding s0 by blast
        have "\<exists> j ni nm nk k \<sigma>i. ?aim k 0 us j nm ni nk fin \<sigma>i" by (rule exI[of _ ?j],rule exI[of _ ?ni],rule exI[of _ 0],
         rule exI[of _ nk],rule exI[of _ k],rule exI[of _ ?\<sigma>i], insert aim, blast)
       }
       with same alts have "\<exists> j ni nm nk k \<sigma>i. ?aim k 0 us j nm ni nk fin \<sigma>i" by fast
       thus ?thesis unfolding s0 by presburger
    qed
   qed
  qed
 qed
qed

lemma ur_sequence1:
 assumes seq:"(\<forall>i<(length us) - 1. (us ! i, us ! Suc i) \<in> par_rstep UR)"
 and no_last_rule:"\<forall>i. (start \<le> i \<and> i < fin) \<longrightarrow> \<not> (\<exists> \<tau> . 
  (us ! i, us ! Suc i) \<in> rstep_r_p_s UR (lhs_n \<rho> (length (snd \<rho>)), rhs_n \<rho> (length (snd \<rho>))) \<epsilon> \<tau>)"
 and rhoR:"\<rho> \<in> R" "length (snd \<rho>) > 0"
 and uroot:"us ! start = Fun Um ws" "U \<rho> m = (More Um Nil Hole (map Var (Z \<rho> m)))" "m < length (snd \<rho>)"
 and sl:"start \<le> fin \<and> fin < length us"
 and nonlv:"non_LV"
 and len:"length us > 0"
 and last_rule:"(us ! fin, us ! (Suc fin)) \<in> rstep_r_p_s UR (lhs_n \<rho> (length (snd \<rho>)), rhs_n \<rho> (length (snd \<rho>))) \<epsilon> \<tau>"
 shows "\<exists> j ni nm \<sigma>i. j (length (snd \<rho>)) = fin \<and> 
 (\<forall>i. m < i \<and> i \<le> length (snd \<rho>) \<longrightarrow> j i < Suc fin \<and> 
  ((us ! (j i), us ! Suc (j i)) \<in> (rstep_r_p_s UR (lhs_n \<rho> i, rhs_n \<rho> i) \<epsilon> (\<sigma>i i)))) \<and> 
 (us ! start, us ! (j (Suc m))) \<in>  par_urstep_below_root ^^ nm \<and>
 (\<forall>i. m < i \<and> i < length (snd \<rho>) \<longrightarrow> (us ! Suc (j i), us ! (j (Suc i))) \<in> par_urstep_below_root ^^ (ni i)) \<and>
 start + nm + (listsum (map (Suc \<circ> ni) [Suc m..<(length (snd \<rho>))])) + 1 = Suc fin"
proof-
 from PairE obtain l r cs where rho:"\<rho> = ((l,r), cs)" by metis
 with rhoR gr0_conv_Suc obtain k' where lcs:"length cs = Suc k'" by auto
 from last_rule[unfolded rstep_r_p_s_def rho snd_conv lcs lhs_n.simps] rho have Uk:"us ! fin = (U \<rho> k')\<langle>ti \<rho>  k'\<rangle> \<cdot> \<tau>" by fastforce
 with U_cond[rule_format, OF rhoR(1), of k'] lcs rho obtain Uk' where Uk':"U \<rho> k' = More Uk' [] \<box> (map Var (Z \<rho> k'))" by fastforce
   let ?no_last = "\<lambda> s us f. \<forall>i. (s \<le> i \<and> i < f) \<longrightarrow> \<not> (\<exists> \<tau> . 
     (us ! i, us ! Suc i) \<in> rstep_r_p_s UR (lhs_n \<rho> (length (snd \<rho>)), rhs_n \<rho> (length (snd \<rho>))) \<epsilon> \<tau>)"
   let ?same = "\<lambda> s f. (us ! s, us ! f) \<in> par_urstep_below_root ^^ (f - s) "
   let ?ex_k = "\<lambda> k f. m < k \<and> k < length (snd \<rho>) \<and> (\<exists> ws' Uk. us ! f = Fun Uk ws' \<and> U \<rho> k = (More Uk Nil Hole (map Var (Z \<rho> k))))" 
   let ?step = "\<lambda> ts ji i \<sigma> . (ts ! ji, ts ! Suc ji) \<in> (rstep_r_p_s UR (lhs_n \<rho> i, rhs_n \<rho> i) \<epsilon> \<sigma> )"
   let ?ur_steps = "\<lambda> k s j f \<sigma>i. \<forall>i. (m < i \<and> i \<le> k \<longrightarrow> s \<le> j i \<and> j i < f \<and> ?step us (j i) i (\<sigma>i i))"                             
   let ?pstep0 = "\<lambda> s j nm. (us ! s, us ! (j (Suc m))) \<in>  par_urstep_below_root ^^ nm"                       
   let ?psteps = "\<lambda> k j ni. (\<forall>i. m < i \<and> i < k \<longrightarrow> (us ! Suc (j i), us ! (j (Suc i))) \<in> par_urstep_below_root ^^ (ni i))" 
   let ?pstepk = "\<lambda> k j nk f. (us ! Suc (j k), us ! f) \<in> par_urstep_below_root ^^ nk"             
   let ?sum = "\<lambda> k st nm ni nk f. st + nm + (listsum (map ni [Suc m..<k])) + (k - m) + nk = f"  
   let ?aim = "\<lambda> k st j nm ni nk f \<sigma>i. (?ex_k k f \<and> ?ur_steps k st j f \<sigma>i \<and> ?pstep0 st j nm \<and> ?psteps k j ni \<and> 
               ?pstepk k j nk f \<and> ?sum k st nm ni nk f)"
 let ?goal2 = "\<lambda>j nm. (us ! start, us ! (j (Suc m))) \<in>  par_urstep_below_root ^^ nm"
 let ?goal1 = "\<lambda>j \<sigma>i. \<forall>i. m < i \<and> i \<le> length (snd \<rho>) \<longrightarrow> j i < Suc fin \<and> 
   (us ! (j i), us ! Suc (j i)) \<in> (rstep_r_p_s UR (lhs_n \<rho> i, rhs_n \<rho> i) \<epsilon> (\<sigma>i i))"
 let ?goal3 = "\<lambda>j ni. \<forall>i. m < i \<and> i < length (snd \<rho>) \<longrightarrow> (us ! Suc (j i), us ! (j (Suc i))) \<in> par_urstep_below_root ^^ (ni i)"
 let ?goal4 = "\<lambda> ni nm. start + nm + (listsum (map (Suc \<circ> ni) [Suc m..<(length (snd \<rho>))])) + 1 = Suc fin"
 from ur_sequence2[OF seq no_last_rule rhoR(1) uroot sl nonlv len] have 
  alts:"(?same start fin) \<or> (\<exists> k ji nm ni nk \<sigma>i. ?aim k start ji nm ni nk fin \<sigma>i)" by blast
 { assume  "\<exists> k ji nm ni nk \<sigma>i. ?aim k start ji nm ni nk fin \<sigma>i"
   then obtain n ji nm ni nk \<sigma>i where dec:"?ex_k n fin" "?ur_steps n start ji fin \<sigma>i" "?pstep0 start ji nm" "?psteps n ji ni"
               "?pstepk n ji nk fin" "?sum n start nm ni nk fin" by blast
   hence k:"m < n" "n < length (snd \<rho>)" "\<exists>ws' Uk. us ! fin = Fun Uk ws' \<and> U \<rho> n = More Uk [] \<box> (map Var (Z \<rho> n))" by auto
   then obtain ws' Un' where Un:"us ! fin = Fun Un' ws'" "U \<rho> n = More Un' [] \<box> (map Var (Z \<rho> n))"  by blast
   from Uk Un Uk' have nk:"Un' = Uk'" by auto
   from U_cond[rule_format, OF rhoR(1) k(2)] Un have "\<forall>\<rho>' n' b c a. (\<rho>' \<in> R \<longrightarrow> n' < length (snd \<rho>') \<longrightarrow>\<not> ((\<rho> \<noteq> \<rho>' \<or> n \<noteq> n') \<and> U \<rho>' n' = More Un' b c a))" by simp
   from this[rule_format, OF rhoR(1), of k', unfolded rho snd_conv lcs, OF lessI] Uk' nk have nk: "n = k'" unfolding rho by fast
   let ?j = "\<lambda>i. if i = length (snd \<rho>) then fin else ji i"
   let ?ni = "\<lambda>i. if i = k' then nk else ni i"
   let ?\<sigma>i = "\<lambda>i. if i = Suc k' then \<tau> else \<sigma>i i"
   from dec(2) last_rule have 1:"?goal1 ?j ?\<sigma>i" unfolding nk rho snd_conv lcs by fastforce
   from dec(3) k(1) have 2:"?goal2 ?j nm" unfolding nk rho snd_conv lcs by auto
   { fix i
     assume " m < i" "i < length (snd \<rho>)"
     with dec(4)[rule_format, of i] dec(5) rho lcs have "(us ! Suc (?j i), us ! (?j (Suc i))) \<in> par_urstep_below_root ^^ (?ni i)" 
      unfolding nk by (cases "i = k'", auto)
   } note 3 = this 
   from k(1) have x:"(map (Suc \<circ> ni) [Suc m..<k']) @ [Suc nk] = map (Suc \<circ> ?ni) [Suc m..<Suc k']" unfolding nk by(induct k', auto)
   with listsum_append have x:"(listsum (map (Suc \<circ> ni) [Suc m..<k'])) + listsum[Suc nk] = listsum (map (Suc \<circ> ?ni) [Suc m..<Suc k'])" by metis
   hence x:"(listsum (map (Suc \<circ> ni) [Suc m..<k'])) + Suc nk = listsum (map (Suc \<circ> ?ni) [Suc m..<Suc k'])" by auto
   from k(1) have "listsum (map ni [Suc m..<k']) + (k' - m) =  listsum (map (Suc \<circ> ni) [Suc m..<k']) + 1" unfolding nk by(induct k', auto)
   hence "start + nm + listsum (map ni [Suc m..<k']) + (k' - m) + nk = start + nm +listsum (map (Suc \<circ> ni) [Suc m..<k']) + Suc nk" by auto
   from dec(6)[unfolded nk this] x have 4:"?goal4 ?ni nm" unfolding rho snd_conv lcs nk by fastforce
   have goal0:"?j (length (snd \<rho>)) = fin" by simp
   with 1 2 3 4 lcs have goal:"?j (Suc k') = fin \<and> ?goal1 ?j ?\<sigma>i \<and> ?goal2 ?j nm \<and> ?goal3 ?j ?ni \<and> ?goal4 ?ni nm" unfolding rho snd_conv lcs by blast
   have "\<exists> j ni nm \<sigma>i. j (length (snd \<rho>)) = fin \<and> ?goal1 j \<sigma>i \<and> ?goal2 j nm \<and> ?goal3 j ni \<and> ?goal4 ni nm" 
    by (rule exI[of _ ?j],rule exI[of _ ?ni],rule exI[of _ nm],rule exI[of _ ?\<sigma>i], insert goal, auto)
   } note case1 = this
   { assume a:"?same start fin"
     from  par_rstep_below_root_pow_same_root[OF this[unfolded uroot]] obtain ys where x:"us ! fin = Fun Um ys" by auto
     with Uk[unfolded Uk'] have Ukm:"Uk' = Um" by simp
     from Uk' U_cond[rule_format, OF rhoR(1), unfolded rho snd_conv lcs, of k']  have "\<forall>\<rho>' n' b c a. (\<rho>' \<in> R \<longrightarrow> n' < length (snd \<rho>') \<longrightarrow>\<not> ((\<rho> \<noteq> \<rho>' \<or> k' \<noteq> n') \<and> U \<rho>' n' = More Uk' b c a))" unfolding rho by auto
     from this[rule_format, OF rhoR(1), of m, unfolded rho snd_conv lcs] uroot have mk: "k' = m" unfolding rho snd_conv lcs Ukm by fast
     let ?j = "\<lambda> i. fin" and ?ni = "\<lambda>i. 0" and ?nm = "fin - start"and ?\<sigma>i = "\<lambda>i. \<tau>"
     from last_rule have 1:"?goal1 ?j ?\<sigma>i" unfolding rho snd_conv lcs mk by (metis le_SucE le_antisym lessI less_imp_le_nat less_not_refl)
     from a have 2:"?goal2 ?j ?nm" by blast
     have 3:"?goal3 ?j ?ni" unfolding rho snd_conv lcs mk by auto
     from sl have 4:"?goal4 ?ni ?nm" unfolding rho snd_conv lcs mk by auto
     from 1 2 3 4 have goal:"?j (length (snd \<rho>)) = fin \<and> ?goal1 ?j ?\<sigma>i \<and> ?goal2 ?j ?nm \<and> ?goal3 ?j ?ni \<and> ?goal4 ?ni ?nm" by blast
     have "\<exists> j ni nm \<sigma>i. j (length (snd \<rho>)) = fin \<and>?goal1 j \<sigma>i \<and> ?goal2 j nm \<and> ?goal3 j ni \<and> ?goal4 ni nm" 
      by (rule exI[of _ ?j],rule exI[of _ ?ni],rule exI[of _ ?nm],rule exI[of _ ?\<sigma>i], insert goal, blast)
    }
    with alts case1 show ?thesis by presburger
qed

lemma linear_term_ti:
 assumes ll:"\<forall> \<rho> \<in> R. ultra_rule left_linear_trs \<rho>"
 and rho:"\<rho> \<in> R"
 and i:"i < length (snd \<rho>)"
 shows "linear_term (ti \<rho> i)"
proof- 
 from ll[rule_format, OF rho, unfolded rules_def] i have lin:"linear_term (lhs_n \<rho> (Suc i))" 
  unfolding left_linear_trs_def by fastforce
 from rho obtain lr cs where rho':"\<rho> = (lr,cs)" by fastforce
 with lin lhs_n.simps(2) have "linear_term (U \<rho> i)\<langle>ti \<rho> i\<rangle>" by auto
 with linear_term.elims(1) U_cond[rule_format, OF rho i] show ?thesis by force
qed

lemma sig_R_ti:
 assumes rho:"\<rho> \<in> R"
 and i:"i < length (snd \<rho>)"
 shows "sig_R (ti \<rho> i)"
proof-  
 obtain lr cs where rho':"\<rho> = (lr, cs)" by fastforce
 with i have set:"(si \<rho> i, ti \<rho> i) \<in> set cs" by fastforce
 have 1:"funs_term (ti \<rho> i) \<subseteq> funs_rule (si \<rho> i, ti \<rho> i)" using funs_rule_def rho' by fastforce
 with set have 2:"funs_term (ti \<rho> i) \<subseteq> funs_trs (set cs)" unfolding funs_trs_def by blast
 with rho 1 2 rho' set show ?thesis unfolding sig_R_def funs_ctrs_def funs_rule_def by force
qed

lemma sig_R_si:
 assumes rho:"\<rho> \<in> R"
 and i:"i < length (snd \<rho>)"
 shows "sig_R (si \<rho> i)"
proof-  
 obtain lr cs where rho':"\<rho> = (lr, cs)" by fastforce
 with i have set:"(si \<rho> i, ti \<rho> i) \<in> set cs" by fastforce
 have 1:"funs_term (si \<rho> i) \<subseteq> funs_rule (si \<rho> i, ti \<rho> i)" using funs_rule_def rho' by fastforce
 with set have 2:"funs_term (si \<rho> i) \<subseteq> funs_trs (set cs)" unfolding funs_trs_def by blast
 with rho 1 2 rho' set show ?thesis unfolding sig_R_def funs_ctrs_def funs_rule_def by force
qed

lemma sig_R_subst: 
 "\<forall>x\<in>vars_term l. sig_R (\<tau> x) \<Longrightarrow> sig_R l \<Longrightarrow> sig_R (l \<cdot> \<tau>)"
proof(induct l)
 case (Var x) 
  from Var(1)[rule_format, unfolded Var] term.simps(17) show ?case by auto
 next
 case (Fun f ts)
  { fix t assume t:"t \<in> set ts"
    then have "vars_term t \<subseteq> vars_term (Fun f ts)" by auto
    with Fun(2) have v:"\<forall>x\<in>vars_term t. sig_R (\<tau> x)" by blast
    from Fun(3)[unfolded sig_R_def funs_ctrs_def term.simps(16)] t have "sig_R t" unfolding sig_R_def funs_ctrs_def by auto
    from Fun(1)[OF t v this] have "sig_R (t \<cdot> \<tau>)" by simp
  } note args = this
  from Fun(3)[unfolded sig_R_def funs_ctrs_def] have "f \<in> funs_ctrs R"  unfolding sig_R_def funs_ctrs_def by force
  with args show ?case unfolding sig_R_def by auto
qed

(* move to Term *)
lemma funs_term_subst: 
 "funs_term (t \<cdot> \<sigma>) = funs_term t \<union> \<Union> ((\<lambda> x. funs_term (\<sigma> x)) ` (vars_term t))"
 by(induct t, auto)

lemma vars_sZi_subset_ti_Z:
 assumes inR:"\<rho> \<in> R" and m:"Suc m < length (snd \<rho>)"
 shows "vars_term (si \<rho> (Suc m)) \<union> set (Z \<rho> (Suc m)) \<subseteq> vars_term (ti \<rho> m) \<union> set (Z \<rho> m)"
proof-
{fix x
 assume x:"x \<in> vars_term (si \<rho> (Suc m))"
 from m length_drop have "length (drop m (snd \<rho>)) > 1" by fastforce
 with nth_drop[of m 1 "snd \<rho>"] m have "snd \<rho> ! (Suc m) \<in> set (drop m (snd \<rho>))"
  by (metis Suc_eq_plus1 nat_less_le nth_mem)
 with m x have "x \<in> vars_trs (set (drop m (snd \<rho>)))" unfolding vars_trs_def using vars_rule_def by fast
 from Z[unfolded Z_vars_def Y_vars_def, rule_format] inR m this have y:"x \<in> set (Y \<rho> m)" by auto
 have " x \<in> vars_term (ti \<rho> m) \<union> set (Z \<rho> m)" proof(cases "x \<in> vars_term (ti \<rho> m)", simp)
  assume x':"x \<notin> vars_term (ti \<rho> m)" 
  from dctrs[rule_format, OF inR m] x have xin:"x \<in> vars_term (clhs \<rho>) \<union> \<Union>((\<lambda>j. vars_term (ti \<rho> j)) ` {j. j < Suc m})" by blast
  have " {j. j < Suc m} = {j. j < m} \<union> {m}" by fastforce
  hence "\<Union>((\<lambda>j. vars_term (ti \<rho> j)) ` {j. j < Suc m}) = (\<Union>((\<lambda>j. vars_term (ti \<rho> j)) ` ({j. j < m} \<union> {m} ))) \<union> vars_term (ti \<rho> m)" by force
  hence "\<Union>((\<lambda>j. vars_term (ti \<rho> j)) ` {j. j < Suc m}) = (\<Union>((\<lambda>j. vars_term (ti \<rho> j)) ` {j. j < m})) \<union> vars_term (ti \<rho> m)" by auto
  with x' xin have xin:"x \<in> vars_term (clhs \<rho>) \<union> \<Union>((\<lambda>j. vars_term (ti \<rho> j)) ` {j. j < m})" by auto
  from m have len:"length (take m (snd \<rho>)) = m" by auto
  with nth_take[of _ m "snd \<rho>"] have "take m (snd \<rho>) = map (\<lambda>j. (snd \<rho> ! j)) [0..< m]"
   by (metis (poly_guards_query) Suc_lessD m take_upt_idx)
  with nth_take[of _ m "snd \<rho>"] len have "set (take m (snd \<rho>)) = (\<lambda>j. (snd \<rho> ! j)) ` {j. j < m}" by auto
  then have aux:"rhss (set (take m (snd \<rho>))) = (\<lambda>j. (ti \<rho> j)) ` {j. j < m}" by force
  from Z[unfolded Z_vars_def] inR m have "X_vars X" by auto
  from this[unfolded X_vars_def, rule_format, OF inR, of m] m xin aux have "x \<in> set (X \<rho> m)" by force
  with y Z[unfolded Z_vars_def] inR m show ?thesis by auto
 qed} note 1 = this
{fix x
 assume x:"x \<in> set (Z \<rho> (Suc m))"
 from Z_imp_X[OF inR m x] X_vars_Suc[OF inR m] have x':"x \<in> set (X \<rho> m) \<union> vars_term (ti \<rho> m)" by auto
 from Z_imp_Y[OF inR m x] Y_vars_mono[OF inR m, of m] have y:"x \<in> set (Y \<rho> m)" by auto
 from Z[unfolded Z_vars_def] inR m x' y have "x \<in> set (Z \<rho> m) \<union> vars_term (ti \<rho> m)" by auto
}
 with 1 show ?thesis by blast
qed

lemma vars_r_subset_tk_Z:
 assumes inR:"((l,r),cs) \<in> R" (is "?\<rho> \<in> R")
 and lcs:"length cs > 0"
 shows  "vars_term r \<subseteq> set (Z ?\<rho> (length cs - 1)) \<union> vars_term (ti ?\<rho> (length cs - 1))"
proof
 fix x
 assume xr:"x \<in> vars_term r"
 let ?k = "length cs - 1"
 from xr type3[rule_format, OF inR] have xx':"x \<in> vars_term l \<union> \<Union>((\<lambda>j. vars_term (ti ?\<rho> j)) ` {j. j < length cs})" by auto
 from lcs have " {j. j < length cs} = {j. j < ?k} \<union> {?k}" by fastforce
  hence "\<Union>((\<lambda>j. vars_term (ti ?\<rho> j)) ` {j. j < length cs}) = (\<Union>((\<lambda>j. vars_term (ti ?\<rho> j)) ` ({j. j < ?k} \<union> {?k} ))) \<union> vars_term (ti ?\<rho> ?k)" by force
 hence "\<Union>((\<lambda>j. vars_term (ti ?\<rho> j)) ` {j. j < length cs}) = (\<Union>((\<lambda>j. vars_term (ti ?\<rho> j)) ` {j. j < ?k})) \<union> vars_term (ti ?\<rho> ?k)" by force
 from xx' this have xx':"x \<in> vars_term l \<union> (\<Union>((\<lambda>j. vars_term (ti ?\<rho> j)) ` {j. j < ?k})) \<union> vars_term (ti ?\<rho> ?k)" by auto
 have len:"length (take ?k (snd ?\<rho>)) = ?k" by auto
 with nth_take[of _ ?k "snd ?\<rho>"] have "take ?k (snd ?\<rho>) = map (\<lambda>j. (snd ?\<rho> ! j)) [0..< ?k]"
  by (metis diff_less lcs snd_conv take_upt_idx zero_less_one)
 hence aux:"rhss (set (take ?k (snd ?\<rho>))) = ((\<lambda>j. (ti ?\<rho> j)) ` {0..< ?k})" by auto
 from xx' conjunct1[OF Z[unfolded Z_vars_def], unfolded X_vars_def, rule_format, OF inR, of ?k] lcs have xx:"x \<in> set (X ?\<rho> ?k) \<union> vars_term (ti ?\<rho> ?k)" unfolding aux by force
 from conjunct1[OF conjunct2[OF Z[unfolded Z_vars_def]], unfolded Y_vars_def, rule_format, OF inR, of ?k] lcs xr this have y:"x \<in> set (Y ?\<rho> ?k)" by auto
 from xx X_Y_imp_Z[OF inR _ _ this] lcs show "x \<in> set (Z ?\<rho> ?k) \<union> vars_term (ti ?\<rho> ?k)" by auto
qed

lemma Lemma_4_2:
 assumes ll:"\<forall> \<rho> \<in> R. ultra_rule left_linear_trs \<rho>"
 and nlv:non_LV
 and s:"sig_R s"
 and t:"sig_R t" "linear_term t"
 and seq:"(s,t\<cdot>\<sigma>) \<in> (par_rstep UR) ^^ n"
 shows "\<exists> \<theta>. (s, t\<cdot>\<theta>) \<in> (cstep R)^* \<and> sig_R_subst \<theta> (vars_term t) \<and> 
 (\<forall> x \<in> vars_term t. (Var x\<cdot>\<theta>, Var x\<cdot>\<sigma>) \<in> (par_rstep UR) ^^ n) \<and> (sig_R (t\<cdot>\<sigma>) \<longrightarrow> t\<cdot>\<theta> = t\<cdot>\<sigma>)"
proof-
 let ?A = "\<lambda> n s t \<sigma> .sig_R s \<and> sig_R t \<and> linear_term t \<and> (s,t \<cdot> \<sigma>) \<in> (par_rstep UR) ^^ n"
 let ?G = "\<lambda> n s t \<sigma> . \<exists> \<theta>. (s, t\<cdot>\<theta>) \<in> (cstep R)^* \<and> sig_R_subst \<theta> (vars_term t) \<and> (\<forall> x \<in> vars_term t. (Var x\<cdot>\<theta>, Var x\<cdot>\<sigma>) \<in> (par_rstep UR) ^^ n) \<and> (sig_R (t\<cdot>\<sigma>) \<longrightarrow> t\<cdot>\<theta> = t\<cdot>\<sigma>)"
 let ?P = "\<lambda> (n,s). \<forall> t \<sigma>. (?A n s t \<sigma> \<longrightarrow> ?G  n s t \<sigma>)"
 have "?P (n,s)" proof(induct rule: wf_induct[OF wf_measures, of "[fst, size \<circ> snd]" ?P])
  case (1 nt) note ind = this
   obtain n s where ns: "nt = (n,s)" by force
   show ?case unfolding ns split proof(rule, rule, rule)
    fix t \<sigma>
    assume a:"sig_R s \<and> sig_R t \<and> linear_term t \<and> (s, t \<cdot> \<sigma>) \<in> par_rstep UR ^^ n"  
    hence sR:"sig_R s" and tR:"sig_R t" and lint:"linear_term t" and seq:"(s, t \<cdot> \<sigma>) \<in> par_rstep UR ^^ n" by auto
    { assume eq:"s = t\<cdot>\<sigma>"
      hence 0:"(s, t\<cdot>\<sigma>) \<in> (cstep R)^*" by auto
      from a have t:"sig_R t" and s:"sig_R s" by auto
      have 1:"sig_R_subst \<sigma> (vars_term t)" proof
       fix x
       assume xt:"x \<in> vars_term t"
       hence "t \<cdot> \<sigma>  \<unrhd> Var x \<cdot> \<sigma>" by (metis supteq_stable vars_term_supteq)
       from funs_term_subterm_subset[OF this] s[unfolded eq] 
        show "sig_R (\<sigma> x)" unfolding sig_R_def funs_ctrs_def aux by auto
      qed
      from par_rstep_refl relpow_refl_mono[OF _ le0 relpow_0_I] have 
        2:"\<forall> x \<in> vars_term t. (Var x\<cdot>\<sigma>, Var x\<cdot>\<sigma>) \<in> (par_rstep UR) ^^ n" by metis
     with 0 1 have " ?G  n s t \<sigma>" by blast
    } note equal_terms = this
    show "?G n s t \<sigma>" proof (cases n)
    (* case 1: the sequence is empty *)
    case 0
      with a equal_terms show ?thesis by auto
    next
    (* case 2: the sequence is non-empty with length Suc m *)
    case (Suc m) note Suc_m = this
     from seq[unfolded Suc relpow_fun_conv] obtain ui where 
      ui:"ui 0 = s" "ui (Suc m) = t \<cdot> \<sigma>" "\<forall>i<Suc m. (ui i, ui (Suc i)) \<in> par_rstep UR" by blast
     show ?thesis proof(cases "\<exists> i < Suc m. \<exists> l r \<sigma>. ui i = l \<cdot> \<sigma> \<and> (ui (Suc i)) = r \<cdot> \<sigma> \<and> (l, r) \<in> UR")
      (* case 2.1: the sequence contains no step at the root *)
      case False note no_root_step = this
       show ?thesis proof (cases "\<exists>x. s = Var x")
       (* case 2.1.1: s = t \<cdot> \<sigma> is a variable *)
        case True
         then obtain x where sx:"s = Var x" by blast
         {fix i
         have "i \<le>  Suc m \<Longrightarrow> ui i = Var x" proof(induct i, insert ui(1) sx, simp)
          case (Suc i)
           hence ix:"ui i = Var x" by auto
           from Suc have "i < Suc m" by auto
           with ix ui(3)[rule_format, OF this, unfolded par_rstep.simps[of "ui i"], unfolded ix] no_root_step 
            show "ui (Suc i) = Var x" by fastforce
          qed}
         from this[OF order_refl]  ui(2) sx have "s = t\<cdot>\<sigma>" by auto
         from equal_terms[OF this] show ?thesis by auto
        next
        (* case 2.1.2: s is not a variable *)
        case False note s_no_var = this 
         from  no_root_step ui(1) have "\<not> (\<exists> l r \<sigma>. s = l \<cdot> \<sigma> \<and> (ui (Suc 0)) = r \<cdot> \<sigma> \<and> (l, r) \<in> UR)" by auto
         with s_no_var ui(3)[rule_format, OF zero_less_Suc, unfolded ui(1) par_rstep.simps[of s]]
          obtain f ss where sf:"s = Fun f ss" by blast
         let ?k = "length ss"
         {fix i
         let ?ps = "\<lambda> ts i. \<forall>x<length ts.(ss ! x, ts ! x) \<in> par_rstep UR ^^ i"
         have "i \<le> Suc m \<Longrightarrow> \<exists> ts. ui i = Fun f ts \<and> ?ps ts i \<and> length ss = length ts" 
          proof(induct i)
          case 0
           from par_rstep_refl relpow_refl_mono[OF _ le0 relpow_0_I] have "?ps ss 0" by metis
           with ui(1)[unfolded sf] exI[of _ ss] show ?case by blast
          next
          case (Suc i)
           hence im:"i \<le> Suc m" "i < Suc m" by auto
            from Suc(1)[OF this(1)] obtain ts where ih:"ui i = Fun f ts" "?ps ts i" "?k = length ts" by blast
           from im no_root_step Suc have "\<not> (\<exists> l r \<sigma>. ui i = l \<cdot> \<sigma> \<and> (ui (Suc i)) = r \<cdot> \<sigma> \<and> (l, r) \<in> UR)" by blast
           with ih(1) ui(3)[rule_format, OF im(2), unfolded par_rstep.simps[of "ui i"]] s_no_var
            obtain ts' where x:"ui (Suc i) = Fun f ts' \<and> (\<forall>x<length ts'. (ts ! x, ts' ! x) \<in> par_rstep UR) \<and> length ts' = length ts" by force
           from conjunct1[OF conjunct2[OF this]] ih(2) have "?ps ts' (Suc i)" by (metis relpow_Suc_I x)
           with Suc ih(3) x show "\<exists>ts. ui (Suc i) = Fun f ts \<and> ?ps ts (Suc i) \<and> ?k = length ts" by force
          qed} note sequence_f_rooted = this
          (* some auxiliaries needed for both of the following cases *)
          from sequence_f_rooted[OF order_refl] ui(2) obtain tss where t\<sigma>:"t \<cdot> \<sigma> = Fun f tss" "length tss =?k" by auto
          from nth_mem supt.arg have ssj:"\<And> j. j < length ss \<Longrightarrow> s \<rhd> ss ! j" unfolding sf by fast
          from supt_size ssj have ms:"\<And> j. j < length ss \<Longrightarrow>((n,ss ! j), nt) \<in> measures [fst, size \<circ> snd]" unfolding ns sf by force
          show ?thesis proof (cases "\<exists> x. t = Var x")
          (* case 2.1.2.1: t is not a variable *)
          case False note t_no_var = this
           from t_no_var subst_apply_term.elims[OF t\<sigma>(1)] t\<sigma>(2) term.inject(2)[of f tss] obtain ts where t:"t = Fun f ts""length ts =?k"
            by (metis length_map)
           with t\<sigma> have sub:"\<forall>x < ?k. tss ! x = ts ! x \<cdot> \<sigma>" by auto
           let ?\<theta>j_cond = "\<lambda> j \<theta>j. (ss ! j, (ts ! j)\<cdot>\<theta>j) \<in> (cstep R)^* \<and> sig_R_subst \<theta>j (vars_term (ts ! j)) \<and> 
            (\<forall> x \<in> vars_term (ts ! j). (Var x\<cdot>\<theta>j, Var x\<cdot>\<sigma>) \<in> (par_rstep UR) ^^ n) \<and> (sig_R ((ts ! j) \<cdot>\<sigma>) \<longrightarrow> (ts ! j)\<cdot>\<theta>j = (ts ! j)\<cdot>\<sigma>)"
           let ?exj = "\<lambda> j. (\<exists> \<theta>j. ?\<theta>j_cond j \<theta>j)"
           have "\<forall>j<length ts. (?exj j)"
           proof(rule, rule)
            fix j
            assume j:"j < length ts"
            let ?tj = "ts ! j" and ?sj = "ss ! j"
            from nth_mem[OF j] supt.arg have ttj:"t \<rhd> ?tj" unfolding t by fast
            from j lint[unfolded t] have lin:"linear_term ?tj" by force
            from sR funs_term_subterm_subset[OF supt_imp_supteq[OF ssj[OF j[unfolded t(2)]]]] 
             have sigs:"sig_R ?sj" unfolding sig_R_def by fast
            from tR funs_term_subterm_subset[OF supt_imp_supteq[OF ttj]] 
             have sigt:"sig_R ?tj" unfolding sig_R_def by fast
            from sub sequence_f_rooted[OF order_refl] ui(2) term.inject(2)[of f tss] t(2) t\<sigma>(1) j
             have step:"(?sj, ?tj \<cdot> \<sigma>) \<in> par_rstep UR ^^ Suc m" by force
            from  ind[rule_format, OF ms[OF j[unfolded t(2)]], unfolded split, rule_format] sigs sigt lin step Suc 
               show "?exj j" by blast
           qed
           from this[unfolded choice_iff'] obtain \<theta>f where \<theta>cond:"\<forall>j<length ts. (?\<theta>j_cond j (\<theta>f j))" by presburger
           let ?\<theta>s = "map \<theta>f [0..<length ts]" and ?vs = "map vars_term ts"
           from lint[unfolded t linear_term.simps(2)] have p:"is_partition ?vs" by auto
           let ?\<theta> = "fun_merge ?\<theta>s ?vs"  
           have xequiv:"\<And> j x . j < length ts \<Longrightarrow> x \<in> vars_term (ts ! j) \<Longrightarrow> ?\<theta> x = (\<theta>f j) x" unfolding length_upt
            proof-
             fix j x
             assume j:"j < length ts" and  x:"x \<in> vars_term (ts ! j)"
             with j have xt:"x \<in> vars_term t" unfolding t(1) by fastforce
             with fun_merge_part[OF p, unfolded length_map, OF j, unfolded nth_map[OF j], OF x, of ?\<theta>s, unfolded nth_map[OF j]
               nth_map[of _ "[0..<length ts]", unfolded length_upt minus_nat.diff_0, OF j]
               nth_upt[of 0 j, unfolded add_0, OF j] in_subst_restrict[OF x]]
               show "?\<theta> x = (\<theta>f j) x" unfolding in_subst_restrict[OF xt] by blast
            qed
            from term_subst_eq this have equiv:"\<And> j. j < length ts \<Longrightarrow> (ts ! j) \<cdot> ?\<theta> = (ts ! j) \<cdot> (\<theta>f j)" unfolding length_upt
            proof-
             fix j
             assume "j < length ts" 
             from term_subst_eq[of "ts ! j" ?\<theta> "\<theta>f j", OF xequiv[OF this]] show "(ts ! j) \<cdot> ?\<theta> = (ts ! j) \<cdot> (\<theta>f j)" by fast
            qed
           let ?\<theta>ts = "map (\<lambda> t. t\<cdot>?\<theta>) ts"
           from t(2) have len:"length ss = length ?\<theta>ts" unfolding length_map by auto
           have "\<forall>j<length ss. (ss ! j, ?\<theta>ts ! j) \<in> (cstep R)\<^sup>*"
            proof(rule,rule)
             fix j 
             assume j:"j < length ss"
             with t(2) have j:"j < length ts" by auto
             from \<theta>cond[rule_format, OF j] equiv[OF j] have "(ss ! j, ts ! j \<cdot> \<theta>f j) \<in> (cstep R)\<^sup>*" by blast
             with equiv[OF j] show "(ss ! j, ?\<theta>ts ! j) \<in> (cstep R)\<^sup>*" unfolding nth_map[OF j] by fastforce
            qed
           from args_steps_imp_steps[OF cstep_ctxt_closed len this]
            have 0:"(s, t\<cdot>?\<theta>) \<in> (cstep R)^*" unfolding sf t by force 
           have 12:"\<forall> x \<in> (vars_term t). (Var x\<cdot>?\<theta>, Var x\<cdot>\<sigma>) \<in> (par_rstep UR) ^^ n \<and> sig_R (?\<theta> x)"
            proof
             fix x
             assume xt:"x \<in> vars_term t"
             from var_imp_var_of_arg[OF this[unfolded t(1)]] obtain j where j:"j < length ts" "x \<in> vars_term (ts ! j)" by force
             from \<theta>cond[rule_format, OF j(1)] j(2) xequiv[OF j(1) j(2)] show "(Var x\<cdot>?\<theta>, Var x\<cdot>\<sigma>) \<in> (par_rstep UR) ^^ n \<and> sig_R (?\<theta> x)" by auto
           qed
           {assume tR:"sig_R (t \<cdot>\<sigma>)"
            { fix j 
              assume j':"j < length ts"
              with t\<sigma>(2) t(2) have j:"j < length tss" by auto
              from tR[unfolded t\<sigma>] funs_term_subterm_subset[OF supt_imp_supteq, OF supt.arg[OF nth_mem[OF j]]] have
               "sig_R (tss ! j)" unfolding sig_R_def by fast
              with \<theta>cond[rule_format, OF j'] sub[rule_format,OF j'[unfolded t(2)]]  equiv[OF j'] 
               have "ts ! j \<cdot>?\<theta> = ts ! j \<cdot> \<sigma>" by presburger
            } 
            with nth_map_conv[OF refl, of ts] t(2) have "t\<cdot>?\<theta> = t\<cdot>\<sigma>" unfolding t(1) subst_apply_term.simps(2) term.simps(2)
             by (metis (poly_guards_query, lifting))
           } note 3 = this
           show ?thesis by (rule exI[of _ ?\<theta>], insert 0 12 3, auto)
         next
          (* case 2.1.2.2: t is a variable *)
          assume "\<exists>x. t = Var x"
          with t\<sigma> obtain x where x:"t = Var x""\<sigma> x = Fun f tss" by force 
          let ?xs = "fresh (length tss)"
          from fresh_vars[rule_format, of "length tss"] have fresh:"distinct ?xs""length ?xs = length tss" by auto
          let ?\<tau> = "mk_subst Var (zip ?xs tss)"
          { fix j assume "j < length tss"
            from mk_subst_distinct[OF fresh(1), unfolded fresh(2), OF refl this] have equiv:"?\<tau> (?xs ! j) = tss ! j" by fast
          } note \<tau>equiv = this
          with fresh(2) nth_map have 
           t\<sigma>':"t\<cdot>\<sigma> = Fun f (map ?\<tau> ?xs)" unfolding t\<sigma> term.inject(2)[of f tss] list_eq_iff_nth_eq length_map by simp
          let ?\<theta>j_cond = "\<lambda> j \<theta>j. (ss ! j, \<theta>j (?xs ! j)) \<in> (cstep R)^* \<and> sig_R_subst \<theta>j {?xs ! j} \<and> 
             (\<theta>j (?xs ! j), ?\<tau> (?xs ! j)) \<in> (par_rstep UR) ^^ n \<and> (sig_R (?\<tau> (?xs ! j)) \<longrightarrow> \<theta>j (?xs ! j) = ?\<tau> (?xs ! j))"
          let ?exj = "\<lambda> j. (\<exists> \<theta>j. ?\<theta>j_cond j \<theta>j)" 
          { fix j 
            assume j:"j < length tss"
            let ?sj = "ss ! j"
            let ?x = "?xs ! j"
            from \<tau>equiv[OF j] sequence_f_rooted[OF order_refl, unfolded ui(2)] t\<sigma> term.inject(2)[of f tss] j have 
             0:"(ss ! j,?\<tau> ?x) \<in> par_rstep UR ^^ Suc m" by simp
            have 1:"linear_term (Var ?x)" by simp
            from sR funs_term_subterm_subset[OF supt_imp_supteq[OF ssj[OF j[unfolded t\<sigma>(2)]]]] 
             have 2:"sig_R ?sj""sig_R (Var ?x)" unfolding sig_R_def by auto
            from ind[rule_format, OF ms[OF j[unfolded t\<sigma>(2)]], unfolded split, rule_format, of "Var ?x" ?\<tau>] 0 1 2 have "?exj j" 
              unfolding term.simps Ball_def singleton_iff simp_thms(41) subst_apply_term.simps(1) Suc by blast
           } 
           hence "\<forall> j < length tss. ?exj j" by auto
           from this[unfolded choice_iff'] obtain \<theta>f where \<theta>cond:"\<forall>j<length tss. (?\<theta>j_cond j (\<theta>f j))" by presburger
           let ?\<theta>s = "map \<theta>f [0..<length tss]" and ?vs = "map (\<lambda> x. {x}) ?xs"
           have p:"is_partition ?vs" unfolding is_partition_def length_map fresh(2) 
            proof(rule, rule, rule, rule)
             fix i j
             assume j:"j < length tss" and i:"i < j"
             hence i':"i < length tss" by auto
             from i fresh(1)[unfolded distinct_conv_nth, rule_format, unfolded fresh(2), OF j i'] have "?xs ! i \<noteq> ?xs ! j" by auto
             thus "map (\<lambda>x. {x}) ?xs ! i \<inter> map (\<lambda>x. {x}) ?xs ! j = {}" 
              unfolding nth_map[of i ?xs, unfolded fresh(2), OF i'] nth_map[of j ?xs, unfolded fresh(2), OF j] by fast
            qed
           def \<theta>' \<equiv> "fun_merge ?\<theta>s ?vs"  
           have xequiv:"\<And> j. j < length tss \<Longrightarrow> \<theta>' (?xs ! j) = (\<theta>f j) (?xs ! j)"
            proof-
             fix j
             assume j:"j < length tss"
             from fun_merge_part[OF p, unfolded length_map fresh(2), OF j, unfolded nth_map[of _ ?xs, unfolded fresh(2), OF j] singleton_iff, of _ ?\<theta>s,
              unfolded nth_map[of _ "[0..<length tss]", unfolded length_upt minus_nat.diff_0, OF j]] nth_upt[of 0 j, unfolded add_0, OF j]  
              show "\<theta>'(?xs ! j) = (\<theta>f j) (?xs ! j)" unfolding \<theta>'_def by auto
            qed
           let ?\<theta>ts = "map \<theta>' ?xs"
          from t\<sigma>(2) fresh(2) have len:"length ss = length ?\<theta>ts" unfolding length_map by auto
          have "\<forall>j<length ss. (ss ! j, ?\<theta>ts ! j) \<in> (cstep R)\<^sup>*"
           proof(rule,rule)
            fix j 
            assume "j < length ss"
            with t\<sigma>(2) have j:"j < length tss" by auto
            from \<theta>cond[rule_format, OF j, unfolded xequiv[OF j, symmetric]]
             show "(ss ! j, ?\<theta>ts ! j) \<in> (cstep R)\<^sup>*" unfolding  nth_map[of j ?xs, unfolded fresh(2), OF j]  by blast
           qed
          from args_steps_imp_steps[OF cstep_ctxt_closed len this]
            have 0:"(s, Fun f ?\<theta>ts) \<in> (cstep R)^*" unfolding sf t by force  
          let ?\<theta> = "subst x (Fun f ?\<theta>ts)"
          have z:"Var x \<cdot> ?\<theta> =  Fun f ?\<theta>ts" by auto
          from 0 have 0:"(s, Var x \<cdot> ?\<theta>) \<in> (cstep R)^*" by fastforce
          { fix j
            assume j:"j < length tss"
            with fresh(2)  have j':"j < length ?xs" by auto
            from \<theta>cond[rule_format, OF j] xequiv[OF j] 
             have "(?\<theta>ts ! j, ?\<tau> (?xs ! j))  \<in> (par_rstep UR) ^^ n \<and> sig_R (?\<theta>ts ! j)" unfolding  nth_map[OF j'] by simp
          } note aux = this
          hence "\<forall> j < length tss.(?\<theta>ts ! j, ?\<tau> (?xs ! j))  \<in> (par_rstep UR) ^^ n" by auto
          with args_par_rstep_pow_imp_par_rstep_pow[of ?\<theta>ts "map ?\<tau> ?xs", unfolded length_map fresh(2)]
            have 1:"(Var x\<cdot>?\<theta>, Var x\<cdot>\<sigma>) \<in> (par_rstep UR) ^^ n" unfolding z t\<sigma>'[unfolded x]
            by (metis Term.term.inject(2) \<tau>equiv t\<sigma>' t\<sigma>(1)) 
          from a[unfolded sf sig_R_def] term.simps(16) have fR:"{f} \<subseteq> funs_ctrs R" unfolding funs_ctrs_def by fast
          from aux len t\<sigma>(2) have aux':"\<forall> j < length ?\<theta>ts.  sig_R (?\<theta>ts ! j)" by presburger
          from this[unfolded all_set_conv_all_nth[symmetric, of ?\<theta>ts sig_R]] fR
           sig_R_def have 2:"sig_R (Var x \<cdot> ?\<theta>)" unfolding subst_simps(2) z by auto
          {
           assume tR:"sig_R (\<sigma> x)"
            { fix j 
              assume j:"j < length tss"
              from tR[unfolded x[symmetric] t\<sigma>, unfolded t\<sigma>[unfolded x, unfolded subst_apply_term.simps(1)]] 
                funs_term_subterm_subset[OF supt_imp_supteq, OF supt.arg[OF nth_mem[OF j]]] have
                "sig_R (tss ! j)" unfolding sig_R_def by fast
              with \<theta>cond[rule_format, OF j, unfolded \<tau>equiv[OF j]] xequiv[OF j] have "\<theta>' (?xs ! j) = tss ! j" by auto
             } 
             with fresh(2) have "(Fun f (map \<theta>' ?xs)) = t\<cdot>\<sigma>" 
              unfolding t\<sigma>(1) subst_apply_term.simps(2) term.simps(2) by (metis map_nth_eq_conv)
             with x have "(Fun f (map \<theta>' ?xs)) = Var x \<cdot>\<sigma>" by auto
         } note 3 = this   
         show ?thesis unfolding x term.simps(17) by (rule exI[of _ ?\<theta>], insert 0 1 2 3, simp)
        qed
       qed
       next
      (* case 2.2: the sequence contains a root step *)
      case True note there_is_root_step = this
       def m0 \<equiv> "LEAST i. (i < Suc m \<and>  (\<exists> l r \<sigma>. ui i = l \<cdot> \<sigma> \<and> (ui (Suc i)) = r \<cdot> \<sigma> \<and> (l, r) \<in> UR))"
       have "m0 < Suc m \<and>  (\<exists> l r \<sigma>. ui m0 = l \<cdot> \<sigma> \<and> (ui (Suc m0)) = r \<cdot> \<sigma> \<and> (l, r) \<in> UR)" unfolding m0_def by (rule LeastI_ex, insert True, auto)
       then obtain l r \<sigma>1 where i:"m0 < Suc m" "ui m0 = l \<cdot> \<sigma>1" "ui (Suc m0) = r \<cdot> \<sigma>1" "(l, r) \<in> UR" by auto
       have first_parsteps:"\<forall> i' < m0. (ui i', ui (Suc i')) \<in> par_urstep_below_root" proof (rule, rule)
        fix i' 
        assume a:"i' < m0" 
        with i(1) ui(3)[rule_format, of i'] have step:"(ui i', ui (Suc i')) \<in> par_rstep UR" by fastforce
        from i(2) nlv[unfolded non_LV_def, rule_format, OF i(4)] obtain g ws where tii:"ui m0 = Fun g ws" by auto
        from ui(3) i(1) have "\<forall> j < Suc m0 - 1. (ui j, ui (Suc j)) \<in> par_rstep UR" by auto
        with no_vars_in_seq'[OF _ nlv, of "Suc m0" ui _ _ i'] tii ui(3) i(1) a have f:"is_Fun (ui i')" by auto
        from not_less_Least[OF a[unfolded m0_def]] a i(1) have "\<not>(\<exists> l r \<sigma>. ui i' = l \<cdot> \<sigma> \<and> (ui (Suc i')) = r \<cdot> \<sigma> \<and> (l, r) \<in> UR)" by force
        with f step[unfolded par_rstep.simps[of "ui i'"]] show "(ui i', ui (Suc i')) \<in> par_urstep_below_root" 
         unfolding  par_urstep_below_root_def is_Var_def by blast
       qed 
       have sl\<sigma>:"(s, l \<cdot> \<sigma>1) \<in> par_urstep_below_root ^^ m0" unfolding relpow_fun_conv[of "ui 0"] ui(1)[symmetric] i(2)[symmetric]
        by (rule exI[of _ ui], insert first_parsteps, auto)
       from all_ctxt_closed_rel_pow[OF all_ctxt_closed_par_rstep] have acc:"\<And> i. all_ctxt_closed UNIV (par_rstep UR ^^ i)" by auto
       show ?thesis proof(cases  "((l, r), []) \<in> R")
        (* case 2.2.1: the first root step applies a (non-conditional) rule which is also in R *)
        case True
         let ?\<rho> = "((l,r),[])"
         let ?cond = "\<lambda> \<theta> s t \<sigma> i . (s, t \<cdot> \<theta>) \<in> (cstep R)\<^sup>* \<and> (\<forall>x\<in>vars_term t. sig_R (\<theta> x)) \<and> 
            (\<forall>x\<in>vars_term t. (Var x \<cdot> \<theta>, Var x \<cdot> \<sigma>) \<in> par_rstep UR ^^ i) \<and> (sig_R (t \<cdot> \<sigma>) \<longrightarrow> t \<cdot> \<theta> = t \<cdot> \<sigma>)"
         (* first part *)
         from ui i(1) i(2) have pre:"(s, l \<cdot> \<sigma>1) \<in> par_rstep UR ^^ m0" unfolding relpow_fun_conv by fastforce
         from True funs_rule_def have 1:"sig_R l \<and> sig_R r" unfolding sig_R_def funs_ctrs_def by force
         from i(4) have "(l,r) \<in> rules ?\<rho>" unfolding rules_def by auto
         with ll[rule_format, OF True] True have 2:"linear_term l" unfolding left_linear_trs_def by fast
         from i(1) have "((m0,s), nt) \<in> measures [fst, size \<circ> snd]" unfolding ns Suc(1) by auto
         from ind[rule_format, OF this, unfolded split] a 1 2 pre obtain \<tau>' where prefix:"?cond \<tau>' s l \<sigma>1 m0" by force
         have "(l \<cdot> \<tau>', r \<cdot> \<tau>') \<in> (cstep_n R (Suc 0))" by(rule cstep_n_SucI[of l r _ _ _ _ _ Hole], insert True, auto)
         hence "(l \<cdot> \<tau>', r \<cdot> \<tau>') \<in> (cstep R)"  unfolding  cstep_def by blast
         with prefix have prefix':"(s, r \<cdot> \<tau>') \<in> (cstep R)\<^sup>*" by force
         (* middle part *)
         from prefix type3[rule_format, OF True, simplified] 
           have "(\<forall>x\<in>vars_term r. (Var x \<cdot> \<tau>', Var x \<cdot> \<sigma>1) \<in> par_rstep UR ^^ m0)" by fast
         with all_ctxt_closed_subst_step[OF acc] have x:"(r \<cdot> \<tau>', r \<cdot> \<sigma>1) \<in> par_rstep UR ^^ m0" by simp
         from prefix type3[rule_format, OF True, simplified] have "(\<forall>x\<in>vars_term r. sig_R (\<tau>' x))" by auto
         with sig_R_subst 1 have 0:"sig_R (r \<cdot> \<tau>')" by presburger
         (* last part *)
         from i(1) obtain j where mij:"Suc m = Suc m0 + j" by (metis add_Suc_right add_Suc_shift less_iff_Suc_add)
         with ui(3) have post:"(ui (Suc m0), ui (Suc m)) \<in> par_rstep UR ^^ j" 
          by(induct j arbitrary: m, insert relpow_Suc_I, auto)
         from relcompI[OF x this[unfolded ui(2) i(3)]] have 1: "(r \<cdot> \<tau>', t \<cdot> \<sigma>) \<in> par_rstep UR ^^ (m0 + j)" unfolding relpow_add by auto
         (* combine *)
         from mij have "((m0 + j, r \<cdot> \<tau>'), nt) \<in> measures [fst, size \<circ> snd]" unfolding ns Suc(1) by auto
         from ind[rule_format, OF this, unfolded split, rule_format] 0 1 a obtain \<theta> where 
           x:"?cond \<theta> (r \<cdot> \<tau>') t \<sigma> (m0 + j)" by presburger
         with mij relpow_refl_mono[of "par_rstep UR" "m0+j" "Suc (m0+j)", OF _ le_SucI[OF le_refl]] par_rstep_refl
           have  "\<forall>x\<in>vars_term t. (Var x \<cdot> \<theta>, Var x \<cdot> \<sigma>) \<in> par_rstep UR ^^ (Suc m)" by fastforce
         with prefix' x have "?cond \<theta> s t \<sigma> (Suc m)" unfolding mij by force 
         thus ?thesis unfolding Suc by (rule exI[of _ \<theta>])
        next
        (* case 2.2.2: the first root step applies a rule which is not in R *)
        case False
         from nlv[unfolded non_LV_def, rule_format, OF i(4), unfolded split] i(2) obtain f us where tif:"ui m0 = Fun f us" by fastforce
         { fix j 
           assume "j \<le> m0"
           then obtain k  where "j + k = m0" by (metis ordered_cancel_comm_monoid_diff_class.le_iff_add)
           hence "\<exists> vs. ui j = Fun f vs" proof(induct k arbitrary: j)
            case 0
             with tif show ?case by auto
            next
            case (Suc k) 
             with i(1) have j:"j < m0" "j < Suc m" by auto
             from Suc(2) add_Suc_shift have "Suc j + k = m0" by simp
             from Suc(1)[OF this] obtain vs where suc:"ui (Suc j) = Fun f vs" by auto
             from not_less_Least[OF j(1)[unfolded m0_def]] j(2) have " \<not> (\<exists>l r \<sigma>. ui j = l \<cdot> \<sigma> \<and> ui (Suc j) = r \<cdot> \<sigma> \<and> (l, r) \<in> UR)" by auto
             with suc ui(3)[rule_format, OF j(2), unfolded par_rstep.simps[of "ui j"]] show ?case by fastforce
           qed
         } 
         with ui(1) obtain ss where s:"s = Fun f ss" by fast
         with a[unfolded sig_R_def] have fR:"f \<in> funs_ctrs R" unfolding funs_ctrs_def by auto
        from i have step_i:"(ui m0, ui (Suc m0)) \<in> rstep_r_p_s UR (l,r) \<epsilon> \<sigma>1" unfolding rstep_r_p_s_def by force
        from ur_step_exhaust[OF this[unfolded tif] nlv] fR obtain \<rho> where exrho:"\<rho> \<in> R" "l = lhs_n \<rho> 0" "r = rhs_n \<rho> 0" by auto
        from exrho(2) PairE lhs_n.simps(1) obtain rr cs where rho:"\<rho> = ((l,rr), cs)" by (metis fst_conv)
        from exrho(3)[unfolded rho rhs_n.simps] False exrho(1)[unfolded rho] have lcs:"0 < length cs" by force
        with rho exrho(3)[unfolded rho rhs_n.simps] have r:"r = (U \<rho> 0)\<langle>fst (cs ! 0)\<rangle>" by force
        let ?last_ur_step = "\<lambda> j' \<tau> . j' > m0 \<and> j' \<le> m \<and> (ui j', ui (Suc j')) \<in> rstep_r_p_s UR (lhs_n \<rho> (length cs),rhs_n \<rho> (length cs)) \<epsilon> \<tau>"
        def usq \<equiv> "map ui [0..<Suc (Suc m)]"
        have lts:"length usq = Suc (Suc m)" unfolding usq_def by force
        from nth_map_upt[of _ "Suc (Suc m)" 0 ui] have uk:"\<And>k. k < Suc (Suc m) \<Longrightarrow> ui k = usq ! k" unfolding usq_def by force
        from this[of "Suc m0"] i(1) i(3) r have ts1:"usq ! (Suc m0) = (U \<rho> 0)\<langle>fst (cs ! 0)\<rangle> \<cdot> \<sigma>1" by simp
        with U_cond[rule_format, OF exrho(1), unfolded rho snd_conv, OF lcs] rho 
           obtain U0 where U0:"U \<rho> 0 = More U0 [] \<box> (map Var (Z \<rho> 0))" by auto
        from ts1[unfolded this] obtain zs where ts1:"usq ! (Suc m0) = Fun U0 zs" by auto
        from ui(3) uk lts have steps:"\<forall>k < length usq - 1. (usq ! k, usq ! (Suc k)) \<in> par_rstep UR" by auto
        (* \<theta>1 can be obtained independent of the below case distinction *) 
        let ?domi = "\<lambda>i. case i of 0 \<Rightarrow> vars_term l | Suc j \<Rightarrow> vars_term (U \<rho> j)\<langle>ti \<rho> j\<rangle>" (* domain of (\<theta>s ! i) *)
        from first_parsteps ui(1) i(2) have 1:"(s, l \<cdot> \<sigma>1) \<in> par_urstep_below_root ^^ m0" unfolding  relpow_fun_conv by blast
        with ll[rule_format, OF exrho(1), unfolded linear_trs_def rules_def rho] lhs_n.simps[of "(l,rr)"] rho 
        have 2:"linear_term l" unfolding left_linear_trs_def by force
        from exrho[unfolded rho] have 3:"sig_R l" unfolding sig_R_def funs_ctrs_def funs_rule_def by force
        from i(1) ns Suc_m have "((m0, s), nt) \<in> measures [fst, size \<circ> snd]" unfolding ns by auto
        from ind[rule_format, OF this, unfolded split, rule_format, of l \<sigma>1] 1 2 3 sR obtain \<theta>1 where
          theta1:"(s, l \<cdot> \<theta>1) \<in> (cstep R)\<^sup>*" "sig_R_subst \<theta>1 (vars_term l)"
             "(\<forall>x\<in>vars_term l. (Var x \<cdot> \<theta>1, Var x \<cdot> \<sigma>1) \<in> par_rstep UR ^^ m0)" "(sig_R (l \<cdot> \<sigma>1) \<longrightarrow> l \<cdot> \<theta>1 = l \<cdot> \<sigma>1)" 
          using relpow_mono[OF par_urstep_par_rstep] by blast
        from theta1(2) l_vars_X0_vars[OF exrho(1)] lcs rho have substR1:"sig_R_subst \<theta>1 (?domi 0)" by auto
        from dctrs[rule_format, OF exrho(1), of 0] lcs rho have vs:"vars_term (si \<rho> 0) \<subseteq> vars_term l" by simp
        with vs theta1(3) have "\<And> x. x \<in> vars_term (si \<rho> 0) \<Longrightarrow>  (Var x \<cdot> \<theta>1, Var x \<cdot> \<sigma>1) \<in> par_rstep UR ^^ m0" by auto
        with all_ctxt_closed_subst_step[OF acc] have sstep:"(si \<rho> 0 \<cdot> \<theta>1, si \<rho> 0 \<cdot> \<sigma>1) \<in> par_rstep UR ^^ m0" by fastforce
        from l_vars_X0_vars[OF exrho(1)] Z_imp_X[OF exrho(1)] lcs rho have vz:"set(Z \<rho> 0) \<subseteq> vars_term l" by auto
        from uk[of m0] i(3) rho uk `m0 < Suc m` r have u1:"usq ! (Suc m0) = (U \<rho> 0)\<langle>si \<rho> 0\<rangle> \<cdot> \<sigma>1" by auto
        (* some abbreviations *)
        let ?ustep = "\<lambda> s t k \<sigma> . (s, t) \<in> rstep_r_p_s UR (lhs_n \<rho> k, rhs_n \<rho> k) \<epsilon> \<sigma>"
        let ?usteps' = "\<lambda> ji. (\<forall>x. \<exists>\<sigma>. 0 < x \<and> x \<le> length cs \<longrightarrow> ?ustep (usq ! ji x) (usq ! Suc (ji x)) x \<sigma>)"
        let ?usteps = "\<lambda> cs \<sigma>i ji. (\<forall>x. 0 < x \<and> x \<le> length cs \<longrightarrow> ?ustep (usq ! ji x) (usq ! Suc (ji x)) x (\<sigma>i x))"
        let ?bef = "\<lambda> ji n1. (usq ! (Suc m0), usq ! ji (Suc 0)) \<in> par_urstep_below_root ^^ n1"
        let ?isteps = "\<lambda> cs ji ni. \<forall>k. 0 < k \<and> k < length cs \<longrightarrow> (usq ! Suc (ji k), usq ! ji (Suc k)) \<in> par_urstep_below_root ^^ ni k"
        show ?thesis proof(cases "\<exists> j' \<tau> . ?last_ur_step j' \<tau>")
        (* case 2.2.2.1: not all term in ui are rooted by U \<rho> i symbols *)
        case True 
         def j \<equiv> "Suc (LEAST j.  \<exists> \<tau>. ?last_ur_step j \<tau>)"
         from LeastI_ex[of "\<lambda>j. \<exists> \<tau>. ?last_ur_step j \<tau>"] True have "\<exists> \<tau>. ?last_ur_step (j - 1) \<tau>" unfolding j_def by fastforce
         then obtain \<tau> where j:"Suc m0 < j"" j \<le> Suc m" "(ui (j - 1), ui j) \<in> rstep_r_p_s UR (lhs_n \<rho> (length cs),rhs_n \<rho> (length cs)) \<epsilon> \<tau>" by force
         from lcs gr0_conv_Suc obtain k' where lcs_Suc:"length cs = Suc k'" by blast
         from j gr0_conv_Suc obtain j' where j_Suc:"j = Suc j'" by blast
         from j(3)[unfolded rstep_r_p_s_def rho rhs_n.simps] have uij:"ui j = rr \<cdot> \<tau>" by auto
         from j(3)[unfolded rstep_r_p_s_def rho lcs_Suc lhs_n.simps] lcs_Suc rho have uij':"ui (j - 1) = (U \<rho> k')\<langle>ti \<rho> k'\<rangle> \<cdot> \<tau>" by auto
         from j have im:"m0 < Suc m" by auto
         (* decompose sequence *)
         from uk[of j] j uij have usq_j_rr:"usq ! j = rr \<cdot> \<tau>" by auto
         from uk[of j'] j(2) j_Suc uij' have usq_j_Uk:"usq ! j' = (U \<rho> k')\<langle>ti \<rho> k'\<rangle> \<cdot> \<tau>" by auto
         from uk[of j'] uk[of j] j have  last_step:"(usq ! j', usq ! j) \<in> rstep_r_p_s UR (lhs_n \<rho> (length cs),rhs_n \<rho> (length cs)) \<epsilon> \<tau>" unfolding j_Suc by force
          { fix i
            assume a:"Suc m0 \<le> i" "i < j - 1"
            from not_less_Least a(2)[unfolded j_def] j_Suc have " \<not> (\<exists>\<tau>. ?last_ur_step i \<tau>)" by auto
            with i a j(2) uk have " \<not> (\<exists>\<tau>.  (usq ! i, usq ! (Suc i)) \<in> rstep_r_p_s UR (lhs_n \<rho> (length cs),rhs_n \<rho> (length cs)) \<epsilon> \<tau>)" by force
          } note j_min = this
         hence j_min:" \<forall>i. Suc m0 \<le> i \<and> i < j' \<longrightarrow>
        \<not> (\<exists>\<tau>. (usq ! i, usq ! Suc i) \<in> rstep_r_p_s UR (lhs_n ((l, rr), cs) (length cs), rhs_n ((l, rr), cs) (length cs)) \<epsilon> \<tau>)" unfolding j_Suc rho by auto
         let ?ji = "\<lambda> cs ji. (\<forall>x. 0 < x \<and> x \<le> length cs \<longrightarrow> ji x < j)"
         let ?sum = "\<lambda> cs n1 ni. Suc m0 + n1 + listsum (map (Suc \<circ> ni) [Suc 0..<length cs]) + 1 = j" 
         from j j_Suc i lts have l:"Suc m0 \<le> j' \<and> j' < length usq" by force
         from lts have "0 < length usq" by auto
         from ur_sequence1[OF steps _ exrho(1), unfolded rho snd_conv, OF j_min lcs ts1 U0[unfolded rho] lcs l nlv, of \<tau>, OF this last_step[unfolded rho j_Suc]] lts lcs_Suc j_Suc
           have "\<exists> ji ni n0 \<sigma>i. ji (length cs) = j' \<and> ?ji cs ji \<and> ?usteps cs \<sigma>i ji \<and> ?bef ji n0 \<and> ?isteps cs ji ni \<and> ?sum cs n0 ni"
            unfolding j_Suc rho using lts lcs_Suc j_Suc by metis
         then obtain ji n0 ni \<sigma>i where dec:"?ji cs ji" "?usteps cs \<sigma>i ji" "?bef ji n0" "?isteps cs ji ni" "?sum cs n0 ni" "ji (length cs) = j'" unfolding rho by metis
         from dec(2)[rule_format, of "length cs"] dec(6) lcs have last_step:"?ustep (usq ! j') (usq ! (Suc j')) (length cs) (\<sigma>i (length cs))" by blast
         (* now the crucial internal lemma *)
         let ?mi = "\<lambda> i. m0 + n0 + listsum (map ni [1..<i])"
         let ?\<theta>cond = "\<lambda> \<theta>s cs cs'. (length \<theta>s = Suc (length cs) \<and> (\<forall> i < Suc (length cs). sig_R_subst (\<theta>s ! i) (?domi i)) \<and>
                     (s, l \<cdot> (\<theta>s ! 0)) \<in> (cstep R)\<^sup>* \<and> 
                     (\<forall> i < length cs. (si \<rho> i \<cdot> (\<theta>s ! i), ti \<rho> i \<cdot> (\<theta>s ! (Suc i))) \<in> (cstep R)\<^sup>*) \<and> 
                     (\<forall> i < length cs. \<forall> j < length (Z \<rho> i) . (((\<theta>s ! i) (Z \<rho> i ! j), (\<theta>s ! Suc i) (Z \<rho> i ! j)) \<in> (cstep R)^*) \<and>
                                                              ((\<theta>s ! Suc i) (Z \<rho> i ! j), (\<sigma>i (Suc i))  (Z \<rho> i ! j)) \<in> par_rstep UR ^^ ?mi (Suc i)) \<and>
                     (\<forall> i < length cs. (\<forall>x\<in>vars_term (ti \<rho> i). ((\<theta>s ! Suc i) x, (\<sigma>i (Suc i)) x) \<in> par_rstep UR ^^ ?mi (Suc i))))"
         let ?theta_prop_with = "\<lambda> \<sigma> m s t \<theta>' i. (s, t \<cdot> \<theta>') \<in> (cstep R)\<^sup>* \<and> sig_R_subst \<theta>' (vars_term t) \<and>
                  (\<forall>x\<in>vars_term t. (Var x \<cdot> \<theta>', Var x \<cdot> \<sigma>) \<in> par_rstep UR ^^ m) \<and> (sig_R (t \<cdot> \<sigma>) \<longrightarrow> t \<cdot> \<theta>' = t \<cdot> \<sigma>)"
         let ?theta_prop = "\<lambda> s t \<theta>' i. ?theta_prop_with (\<sigma>i i) (?mi i) s t \<theta>' i"
         { fix cs cs'
           assume "0 < length cs" "\<rho> = ((l,rr),cs @ cs')" "?ji cs ji" "?usteps cs \<sigma>i ji" "?bef ji n0" "?isteps cs ji ni"
                  "?sum (cs @ cs') n0 ni"
         hence "\<exists> \<theta>s. ?\<theta>cond \<theta>s cs cs'"
         proof(induct "length cs" arbitrary:cs cs')
          case 0
           thus ?case by simp
          next
           case (Suc k)
            note U_cond[rule_format, OF exrho(1)] = U_cond
            { fix m y
              assume a:"m = Suc y" "m \<le> length cs"
              from Suc(6)[rule_format, of m] this have
               "(usq ! (ji m), usq ! (Suc (ji m))) \<in> rstep_r_p_s UR (lhs_n \<rho> m, rhs_n \<rho> m) \<epsilon> (\<sigma>i m)" by auto
              from this[unfolded rstep_r_p_s_def] have lhs:"usq ! (ji m) = lhs_n \<rho> m \<cdot> (\<sigma>i m)" by force
              from lhs[unfolded Suc(4) a lhs_n.simps] a Suc(4) Suc(3) have "usq ! (ji m) = (U \<rho> y)\<langle>ti \<rho> y\<rangle> \<cdot> (\<sigma>i m)" by auto
            } note lhss_in_seq = this
            show ?case proof (cases "k = 0")
             case True
              note 0 = Suc(3) and rho = Suc(4) and inR = exrho(1)
              let ?t0 = "ti \<rho> 0" and ?\<sigma>2 = "\<sigma>i (Suc 0)"
              from lhss_in_seq[of "Suc 0" 0] rho Suc(2)[unfolded True] have u2:"usq ! (ji (Suc 0)) = (U \<rho> 0)\<langle>?t0\<rangle> \<cdot> ?\<sigma>2" by auto
              from u1 u2 Suc(7) have m0_step:"((U \<rho> 0)\<langle>si \<rho> 0\<rangle> \<cdot> \<sigma>1, (U \<rho> 0)\<langle>?t0\<rangle> \<cdot> ?\<sigma>2) \<in> par_urstep_below_root ^^ n0" by auto
              from this[unfolded U0 ctxt_apply.simps(2)] par_rstep_pow_imp_args_par_rstep_pow 
                  have "(si \<rho> 0 \<cdot> \<sigma>1, ?t0 \<cdot> ?\<sigma>2) \<in> par_rstep UR ^^ n0" by fastforce
              with sstep have sstep:"(si \<rho> 0 \<cdot> \<theta>1, ?t0 \<cdot> ?\<sigma>2) \<in> par_rstep UR ^^ (m0 + n0)" unfolding relpow_add by blast (* #4 *)
              from sig_R_si[OF exrho(1)] 0 rho have "sig_R (si \<rho> 0)" by auto
              with vs theta1(2) have 1:"sig_R ((si \<rho> 0) \<cdot> \<theta>1)" unfolding sig_R_def funs_term_subst by blast
              from linear_term_ti[OF ll exrho(1), unfolded rho, simplified] 0 rho have 2:"linear_term ?t0" by simp
              from sig_R_ti[OF exrho(1)] 0 rho have 3:"sig_R ?t0"  by auto
              from i(1) Suc(9) j(1) j(2) have m:"((m0 + n0, si \<rho> 0 \<cdot> \<theta>1), nt) \<in> measures [fst, size \<circ> snd]" 
               unfolding ns Suc_m by force
              from ind[rule_format, OF m, unfolded split, rule_format, of ?t0 ?\<sigma>2] sstep 1 2 3 obtain \<theta>' where
               r2: "?theta_prop (si \<rho> 0 \<cdot> \<theta>1) ?t0 \<theta>' (Suc 0)" by auto (* #6 *)
               { fix j 
                 assume jl:"j < length (Z \<rho> 0)" 
                 let ?z = "(Z \<rho> 0) ! j" 
                 from vz theta1(3) nth_mem[OF jl] have i:"(Var ?z \<cdot> \<theta>1, Var ?z \<cdot> \<sigma>1) \<in> par_rstep UR ^^ m0" by auto (* #2 *)
                 from m0_step[unfolded U0 ctxt_apply.simps(2)] iffD1[OF par_rstep_pow_imp_args_par_rstep_pow] jl
                  have "(Var ?z \<cdot> \<sigma>1, Var ?z \<cdot> ?\<sigma>2) \<in> par_rstep UR ^^ n0" by fastforce
                 with i have steps:"(Var ?z \<cdot> \<theta>1, Var ?z \<cdot> ?\<sigma>2) \<in> par_rstep UR ^^ (m0 + n0)"  unfolding relpow_add by blast (* #5 *)
                 from vz jl theta1(2) have 1:"sig_R (\<theta>1 ?z)" unfolding sig_R_def funs_term_subst by force
                 have 2:"linear_term (Var ?z)" by auto
                 have 3:"sig_R (Var ?z)" unfolding sig_R_def by auto
                 from i(1) Suc(9) j(1) j(2) have "((m0+n0, \<theta>1 ?z), nt) \<in> measures [fst, size \<circ> snd]" 
                  unfolding ns Suc_m by force
                 from ind[rule_format, OF this, unfolded split, rule_format, of "Var ?z" ?\<sigma>2] steps 1 2 3 have
                   "\<exists> \<theta>z. ?theta_prop (\<theta>1 ?z) (Var ?z) \<theta>z (Suc 0)" by force (* #7 *)
               } note zsteps = this
               with choice[of "\<lambda> j \<theta>z. j < length (Z \<rho> 0) \<longrightarrow>  ?theta_prop (\<theta>1 ((Z \<rho> 0) ! j)) (Var ((Z \<rho> 0) ! j)) \<theta>z (Suc 0)"] 
                 obtain \<theta>z where zsteps:"\<forall>j < length (Z \<rho> 0). ?theta_prop (\<theta>1 ((Z \<rho> 0) ! j)) (Var ((Z \<rho> 0) ! j)) (\<theta>z j) (Suc 0)" by auto
               let ?\<theta>s = "\<theta>' # map \<theta>z [0..<length (Z \<rho> 0)]" and ?vs = "vars_term ?t0 # map (\<lambda> z. {z}) (Z \<rho> 0)"
               from Z_part'[OF inR _ ll[rule_format, OF inR]] 0 rho have p:"is_partition ?vs" by auto
               have l0:"0 < length ?vs" by auto
               def \<theta>2 \<equiv> "fun_merge ?\<theta>s ?vs"
               from fun_merge_part[OF p l0, of _ ?\<theta>s] have vt:"\<And>x. x \<in> vars_term ?t0 \<Longrightarrow> \<theta>' x = \<theta>2 x" unfolding \<theta>2_def by fastforce
               with r2 term_subst_eq[of "ti \<rho> 0" \<theta>' \<theta>2, OF this] have st_prop:"?theta_prop (si \<rho> 0 \<cdot> \<theta>1) ?t0 \<theta>2 1" by simp (* contains #8 *)
               { fix j 
                 let ?z = "(Z \<rho> 0) ! j"
                 assume j:"j < length (Z \<rho> 0)"
                 hence j':"Suc j < length ?vs" by auto 
                 with j have jth:"?vs ! (Suc j) = {?z}" by simp
                 from fun_merge_part[OF p j', unfolded jth, of _ ?\<theta>s] j have "\<theta>z j ?z = \<theta>2 ?z" unfolding \<theta>2_def by simp
                 with zsteps[rule_format, OF j] have p:"?theta_prop (\<theta>1 ?z) (Var ?z) \<theta>2 1" by simp
               } note z_prop = this (* contains #9 *) 
              have a1:"vars_term ((U \<rho> 0)\<langle>ti \<rho> 0\<rangle>) = vars_term (ti \<rho> 0) \<union> set (Z \<rho> 0)" unfolding U0 by simp
              with st_prop z_prop have substR2:"sig_R_subst \<theta>2 (?domi 1)"  unfolding set_conv_nth[of "Z \<rho> 0"] by fastforce
              from r2 all_ctxt_closed_subst_step[OF acc, of ?t0 \<theta>2 ?\<sigma>2] vt have x1:"(?t0 \<cdot> \<theta>2, ?t0 \<cdot> ?\<sigma>2) \<in> par_rstep UR ^^ ?mi 1" by force
              from z_prop have x2:"\<And> j. j < length (Z \<rho> 0) \<longrightarrow>  (\<theta>2 ((Z \<rho> 0) ! j), ?\<sigma>2 ((Z \<rho> 0) ! j)) \<in> par_rstep UR ^^ ?mi 1" by simp
              let ?\<theta>s = "[\<theta>1, \<theta>2]"
              have len:"length ?\<theta>s = 2" by force
              from Suc(2)[unfolded True] have aux:"\<And>P. P \<theta>1 \<theta>2 0 \<Longrightarrow> \<forall>i<length cs. P (?\<theta>s ! i) (?\<theta>s ! (Suc i)) i"
               by (metis less_Suc0 nth_Cons_0 nth_Cons_Suc)
              from Suc(2)[unfolded True] have f1:" length ?\<theta>s = Suc (length cs)" by auto
              from substR1 substR2 have f2:"\<forall> i < Suc (length cs). sig_R_subst (?\<theta>s ! i) (?domi i)" 
               unfolding Suc(2)[unfolded True, symmetric] all_less_two by auto
              have f4:"\<forall>i<length cs. (si \<rho> i \<cdot> ?\<theta>s ! i, ti \<rho> i \<cdot> ?\<theta>s ! Suc i) \<in> (cstep R)\<^sup>*" by (rule aux, insert st_prop, auto)
              have f5:"(\<forall>i<length cs. \<forall>j < length (Z \<rho> i). ((Var (Z \<rho> i ! j) \<cdot> ?\<theta>s ! i, Var (Z \<rho> i ! j) \<cdot>?\<theta>s ! Suc i) \<in> (cstep R)\<^sup>* \<and>
                                                           ((?\<theta>s ! Suc i) (Z \<rho> i ! j), \<sigma>i (Suc i) (Z \<rho> i ! j)) \<in> par_rstep UR ^^ ?mi i))" 
               by (rule aux, insert z_prop, auto) 
              have f6:"\<forall>i<length cs. \<forall>x\<in>vars_term (ti \<rho> i). ((?\<theta>s ! Suc i) x, \<sigma>i (Suc i) x) \<in> par_rstep UR ^^ ?mi i" 
               by (rule aux, insert Suc(2)[unfolded True] st_prop, auto)
              from f1 f2 theta1(1) f4 f5 f6 have c:"?\<theta>cond ?\<theta>s cs cs'" by fastforce
              show ?thesis by (rule exI[of _ ?\<theta>s], insert c, auto)
             next
             case False note k_pos = this
              from not0_implies_Suc[OF this] obtain m where k:"k = Suc m" by blast
              note inR = exrho(1) and rho = Suc(4)
              from k Suc(2) Suc(3) have ml:"m < length cs" and kl:"k < length cs" by auto
              from U_cond[unfolded rho snd_conv length_append, OF trans_less_add1[OF kl]] kl rho obtain Uk where 
               Uk:"U \<rho> k = More Uk [] \<box> (map Var (Z \<rho> k))" by blast
              from U_cond[unfolded rho snd_conv length_append, OF trans_less_add1[OF ml]] ml rho obtain Um where 
               Um:"U \<rho> m = More Um [] \<box> (map Var (Z \<rho> m))" by blast
              let ?cs = "butlast cs" and ?cs' = "last cs # cs'"
              from append_butlast_last_id[of cs] Suc(3) have app:"cs @ cs' = ?cs @ ?cs'" by force
              from length_butlast[of cs] Suc(2) have 0:"length ?cs = k" by auto
              with k have 1:" 0 < length ?cs" by auto
              from Suc(2) app have 2:"\<rho> = ((l,rr), ?cs @ ?cs')" unfolding rho by force
              from Suc(5) have 3:"?ji ?cs ji" by force
              from Suc(6) have 4:"?usteps ?cs \<sigma>i ji" by force
              from Suc(8) have 5:"?isteps ?cs ji ni" by force
              from Suc(9) app have 6:"?sum (?cs @ ?cs') n0 ni" by simp
              from Suc(1)[OF 0[symmetric] 1 2 3 4 Suc(7) 5 6] obtain \<theta>s' where \<theta>s':"?\<theta>cond \<theta>s' ?cs ?cs'" by auto
              let ?k = "Suc k" 
               from \<theta>s' k 0 have vt:"\<forall>x\<in>vars_term (ti \<rho> m). ((\<theta>s' ! k) x, \<sigma>i k x) \<in> par_rstep UR ^^ ?mi k" by auto
               from \<theta>s' k 0 have vz':"\<forall>j < length (Z \<rho> m). ((\<theta>s' ! k) (Z \<rho> m ! j), \<sigma>i k (Z \<rho> m ! j)) \<in> par_rstep UR ^^ ?mi k" by auto
               hence vz:"\<forall>z \<in> set (Z \<rho> m). ((\<theta>s' ! k) z, \<sigma>i k z) \<in> par_rstep UR ^^ ?mi k" unfolding all_set_conv_all_nth by blast
               let ?\<theta>k = "\<theta>s' ! k" and ?\<sigma>k = "\<sigma>i (Suc m)"
               from vars_sZi_subset_ti_Z[OF inR] k rho k Suc(2) have vs:"vars_term (si \<rho> k) \<subseteq> vars_term (ti \<rho> m) \<union> set (Z \<rho> m)" by simp
               with vt vz 0 have "\<forall>x\<in>vars_term (si \<rho> k). ((\<theta>s' ! k) x, \<sigma>i k x) \<in> par_rstep UR ^^ ?mi k" unfolding k by fast
               from this all_ctxt_closed_subst_step[OF acc] 
                have si_k_steps:"((si \<rho> k) \<cdot> (\<theta>s' ! k),(si \<rho> k) \<cdot> (\<sigma>i k)) \<in> par_rstep UR ^^ ?mi k" by fast (* #2 *)
               from k kl Suc(6)[rule_format, of k] have x:"(usq ! ji k, usq ! Suc (ji k)) \<in> rstep_r_p_s UR (lhs_n \<rho> k, rhs_n \<rho> k) \<epsilon> (\<sigma>i k)" by auto
               from rhs_n.simps[of "(l,r)" "cs @ cs'" k] rho Suc(2) have "rhs_n \<rho> k = (U \<rho> k)\<langle>si \<rho> k\<rangle>" by simp 
               with x[unfolded rho k] rho rstep_r_p_s_def Suc(2) have 
                u1:"usq ! Suc (ji k) = (U \<rho> k)\<langle>si \<rho> k\<rangle> \<cdot> (\<sigma>i k)" unfolding k by force
               from Suc(2) Suc(6)[rule_format, of "?k"] have 
                 "(usq ! ji ?k, usq ! Suc (ji ?k)) \<in> rstep_r_p_s UR (lhs_n \<rho> ?k, rhs_n \<rho> ?k) \<epsilon> (\<sigma>i ?k)" by force
               from this[unfolded rho lhs_n.simps(2)] rstep_r_p_s_def rho have
                u2:" usq ! ji ?k = (U \<rho> k)\<langle>ti \<rho> k\<rangle> \<cdot> (\<sigma>i ?k)" by force
               from u1 u2 k kl Suc(8)[rule_format, of k] have 
                u:"((U \<rho> k)\<langle>si \<rho> k\<rangle> \<cdot> (\<sigma>i k),(U \<rho> k)\<langle>ti \<rho> k\<rangle> \<cdot> (\<sigma>i ?k)) \<in> par_urstep_below_root ^^ ni k" by auto
              from this[unfolded Uk ctxt_apply.simps append_Nil] par_urstep_below_root_def par_rstep_pow_imp_args_par_rstep_pow
               have si_k_steps':"(si \<rho> k \<cdot> (\<sigma>i k), ti \<rho> k \<cdot> (\<sigma>i ?k)) \<in> par_rstep UR ^^ ni k" by auto 
              from k kl k_pos have sum:"?mi ?k = ?mi k + ni k" by simp
              from si_k_steps si_k_steps' relpow_add have 
               step:"((si \<rho> k) \<cdot> (\<theta>s' ! k),  ti \<rho> k \<cdot> (\<sigma>i ?k)) \<in> par_rstep UR ^^ ?mi ?k" unfolding sum by fast (* #4 *)
              from sig_R_si[OF inR] kl rho have sig_R:"sig_R (si \<rho> k)" by auto
              from vs have "vars_term (si \<rho> k) \<subseteq> vars_term (U \<rho> m)\<langle>ti \<rho> m\<rangle>" unfolding Um ctxt_apply.simps by simp
              with sig_R conjunct1[OF conjunct2[OF \<theta>s'], rule_format, of k] k 
               have 1:"sig_R ((si \<rho> k) \<cdot> ?\<theta>k)" unfolding sig_R_def funs_term_subst 0 by auto
              from linear_term_ti[OF ll inR, unfolded rho, simplified] kl rho have 2:"linear_term (ti \<rho> k)" by simp
              from sig_R_ti[OF inR] kl rho have 3:"sig_R (ti \<rho> k)"  by auto 
              have "\<And>j. Suc 0 \<le> j \<and> j < Suc k \<Longrightarrow> ni j \<le> (Suc \<circ> ni) j" by force
              have "listsum (map ni [Suc 0..<Suc k]) \<le> listsum (map (Suc \<circ> ni) [Suc 0..<Suc k])" by (induct k, auto)
              with Suc(9) have m':"?mi ?k < j" unfolding length_append Suc(2)[symmetric] 
               upt_add_eq_append[OF Suc_leI[OF Suc(3)], of "length cs'", unfolded Suc(2)[symmetric]] map_append listsum_append by fastforce
              with j(2) have m:"((?mi ?k, si \<rho> k \<cdot> ?\<theta>k), nt) \<in> measures [fst, size \<circ> snd]" unfolding ns Suc_m by auto
              from ind[rule_format, OF m, unfolded split, rule_format] 1 2 3 step obtain \<theta>' where
               \<theta>':"?theta_prop (si \<rho> k \<cdot> ?\<theta>k) (ti \<rho> k) \<theta>' ?k" by blast
              (* ... the case for Zs ... *) 
              { fix j
                assume jl:"j < length (Z \<rho> k)" 
                let ?z = "(Z \<rho> k) ! j" 
                from u[unfolded Uk ctxt_apply.simps append_Nil subst_apply_term.simps(2)] par_rstep_pow_imp_args_par_rstep_pow
                 have "(\<forall>i< length (Z \<rho> k). (map (\<sigma>i k) (Z \<rho> k) ! i, map (\<sigma>i ?k) (Z \<rho> k) ! i) \<in> par_rstep UR ^^ ni k)" by auto
                from this[rule_format, OF jl] nth_map[of j] jl have z_step:"((\<sigma>i k) ?z, \<sigma>i ?k ?z) \<in> par_rstep UR ^^ ni k" by force
                from vars_sZi_subset_ti_Z[OF inR] k rho k Suc(2) have "set (Z \<rho> k) \<subseteq> vars_term (ti \<rho> m) \<union> set (Z \<rho> m)" by auto
                with nth_mem[OF jl] have vz':"?z \<in> vars_term (ti \<rho> m) \<union> set (Z \<rho> m)" by auto
                with vz vt have "((\<theta>s' ! k) ?z, \<sigma>i k ?z) \<in> par_rstep UR ^^ ?mi k" by auto
                with z_step relpow_add have step:"((\<theta>s' ! k) ?z, \<sigma>i ?k ?z) \<in> par_rstep UR ^^ ?mi ?k" unfolding sum by fast (* #5 *)
                have sig_R:"sig_R (Var ?z)" unfolding sig_R_def by simp
                from vz' have "?z \<in> vars_term (U \<rho> m)\<langle>ti \<rho> m\<rangle>" unfolding Um ctxt_apply.simps by simp
                with sig_R conjunct1[OF conjunct2[OF \<theta>s'], rule_format, of k] k 
                 have 1:"sig_R (?\<theta>k ?z)" unfolding sig_R_def funs_term_subst 0 by auto
                from m' j(2) have m:"((?mi ?k, ?\<theta>k ?z), nt) \<in> measures [fst, size \<circ> snd]" unfolding ns Suc_m by auto
               from ind[rule_format, OF m, unfolded split, rule_format, of "Var ?z" "\<sigma>i ?k"] 1 sig_R step have "\<exists> \<theta>z. ?theta_prop (?\<theta>k ?z) (Var ?z) \<theta>z ?k" by auto
              }  with choice[of "\<lambda> j \<theta>z. j < length (Z \<rho> k) \<longrightarrow>  ?theta_prop (?\<theta>k ((Z \<rho> k) ! j)) (Var ((Z \<rho> k) ! j)) \<theta>z ?k"] 
                 obtain \<theta>z where zsteps:"\<forall>j < length (Z \<rho> k). ?theta_prop (?\<theta>k ((Z \<rho> k) ! j)) (Var ((Z \<rho> k) ! j)) (\<theta>z j) ?k" by auto
               let ?\<theta>s = "\<theta>' # map \<theta>z [0..<length (Z \<rho> k)]" and ?vs = "vars_term (ti \<rho> k) # map (\<lambda> z. {z}) (Z \<rho> k)"
               from Z_part'[OF inR _ ll[rule_format, OF inR], of k] Suc(2) rho have p:"is_partition ?vs" by auto
               have l0:"0 < length ?vs" by auto
               def \<theta>Sk \<equiv> "fun_merge ?\<theta>s ?vs"
               from fun_merge_part[OF p l0, of _ ?\<theta>s] have vt:"\<And>x. x \<in> vars_term (ti \<rho> k) \<Longrightarrow> \<theta>' x = \<theta>Sk x" unfolding \<theta>Sk_def by fastforce
               with \<theta>' term_subst_eq[of "ti \<rho> k" \<theta>' \<theta>Sk, OF this] have st_prop:"?theta_prop (si \<rho> k \<cdot> ?\<theta>k) (ti \<rho> k) \<theta>Sk ?k" by simp (* #6 *)
               { fix j 
                 let ?z = "(Z \<rho> k) ! j"
                 assume j:"j < length (Z \<rho> k)"
                 hence j':"Suc j < length ?vs" by auto 
                 with j have jth:"?vs ! (Suc j) = {?z}" by simp
                 from fun_merge_part[OF p j', unfolded jth, of _ ?\<theta>s] j have "\<theta>z j ?z = \<theta>Sk ?z" unfolding \<theta>Sk_def by simp
                 with zsteps[rule_format, OF j] have p:"?theta_prop (?\<theta>k ?z) (Var ?z) \<theta>Sk ?k" by simp
               } note z_prop = this (* #7 *)
              let ?\<theta>s = "\<theta>s' @ [\<theta>Sk]"
              from \<theta>s' 0 Suc(2) have f1:"length ?\<theta>s = Suc ?k" by fastforce
              with 0 Suc(2) last_snoc last_conv_nth[of ?\<theta>s] have last:"?\<theta>s ! ?k = \<theta>Sk" by fastforce
              from f1 0 Suc(2) conjunct1[OF \<theta>s'] nth_append[of \<theta>s' "[\<theta>Sk ]" k] have butlast:"?\<theta>s ! k = ?\<theta>k" by fastforce
              from f1 Suc(2) 0 last have aux:"\<And>P. P ?\<theta>k \<theta>Sk k \<Longrightarrow> (\<forall> i < k. P (\<theta>s' ! i) (\<theta>s' ! Suc i) i) \<Longrightarrow> \<forall>i< ?k. P (?\<theta>s ! i) (?\<theta>s ! (Suc i)) i"
               by (metis (erased, hide_lams) append_Cons_nth_left length_append_singleton less_Suc_eq not_less_eq)
              { fix i
                assume i:"i < Suc ?k"
                have "sig_R_subst (?\<theta>s ! i) (?domi i)" proof(cases "i = ?k")
                 case True
                  have a1:"vars_term ((U \<rho> k)\<langle>ti \<rho> k\<rangle>) = vars_term (ti \<rho> k) \<union> set (Z \<rho> k)" unfolding Uk by simp
                  with st_prop z_prop show ?thesis unfolding set_conv_nth[of "Z \<rho> k"] True last by force
                 next
                 case False
                  with i have "i < ?k" by auto
                  with conjunct1[OF conjunct2[OF \<theta>s'], rule_format, unfolded 0, OF this] nth_append[of \<theta>s' "[\<theta>Sk ]" i] conjunct1[OF \<theta>s'] 0
                   show ?thesis unfolding sig_R_def by auto
                 qed
              } note f2 = this
              have f4:"\<forall>i < ?k. (si \<rho> i \<cdot> ?\<theta>s ! i, ti \<rho> i \<cdot> ?\<theta>s ! Suc i) \<in> (cstep R)\<^sup>*" by (rule aux, insert st_prop \<theta>s' Suc(2), auto)
              from \<theta>s' nth_append[of \<theta>s' "[\<theta>Sk ]" 0] have f3:"(s, l \<cdot> ?\<theta>s ! 0) \<in> (cstep R)\<^sup>*" by auto
              { fix i j 
                assume i:"i< ?k" and j:"j < length (Z \<rho> i)"
                have "(((?\<theta>s ! i) (Z \<rho> i ! j), (?\<theta>s ! Suc i) (Z \<rho> i ! j)) \<in> (cstep R)\<^sup>* \<and>
                      ((?\<theta>s ! Suc i) (Z \<rho> i ! j), \<sigma>i (Suc i) (Z \<rho> i ! j)) \<in> par_rstep UR ^^ ?mi (Suc i))"
                proof(cases "i < k") 
                 case False
                  with i have ik:"i = k" by auto
                  from z_prop[OF j[unfolded ik]] show ?thesis unfolding ik last butlast by auto
                 next
                 case True
                  with conjunct1[OF conjunct2[OF conjunct2[OF conjunct2[OF conjunct2[OF \<theta>s']]]], rule_format, OF True[unfolded 0[symmetric]] j] 
                   nth_append[of \<theta>s' "[\<theta>Sk ]" i] nth_append[of \<theta>s' "[\<theta>Sk ]" "Suc i"] conjunct1[OF \<theta>s', unfolded 0] show ?thesis by force
                qed
               } note f5 = this 
              have f6:"\<forall>i<Suc k. \<forall>x\<in>vars_term (ti \<rho> i). ((?\<theta>s ! Suc i) x, \<sigma>i (Suc i) x) \<in> par_rstep UR ^^ ?mi (Suc i)" 
               by (rule aux, insert Suc(2) \<theta>s' st_prop, auto)
              from f1 f2 f3 f4 f5 f6 have c:"?\<theta>cond ?\<theta>s cs cs'" unfolding Suc(2)[symmetric] by auto
              show "\<exists> \<theta>s. ?\<theta>cond \<theta>s cs cs'" by (rule exI[of _ ?\<theta>s], insert c, auto)
             qed
            qed
           } (* end of crucial internal lemma *) 
           from this[OF lcs, of "[]", unfolded append_Nil2, OF rho dec(1) dec(2) dec(3) dec(4) dec(5)] obtain \<theta>s where \<theta>s:"?\<theta>cond \<theta>s cs []" by presburger 
           (* apply argument once more to r *)
           let ?k = "length (snd \<rho>)"
           from \<theta>s rho have 1:"\<forall>i < ?k. \<forall>z\<in>set (Z \<rho> i). (Var z \<cdot> \<theta>s ! i, Var z \<cdot> \<theta>s ! Suc i) \<in> (cstep R)\<^sup>*" unfolding all_set_conv_all_nth by auto
           with A1[OF exrho(1) ll[rule_format, OF exrho(1)], of \<theta>s] lcs 1 rho \<theta>s have "(l \<cdot> \<theta>s ! 0, rr \<cdot> \<theta>s ! ?k) \<in> (cstep R)\<^sup>+" by auto
           with \<theta>s have ssteps:"(s, rr \<cdot> \<theta>s ! length (snd \<rho>)) \<in> (cstep R)\<^sup>+" by auto
           from lcs gr0_conv_Suc rho obtain k' where k:"?k = Suc k'""length cs = Suc k'" unfolding rho by auto
           from U_cond[unfolded rho snd_conv length_append, rule_format, OF exrho(1), of k'] rho lcs obtain Uk where 
               Uk:"U \<rho> k' = More Uk [] \<box> (map Var (Z \<rho> k'))" unfolding k by force 
           hence vs:"vars_term (U \<rho> k')\<langle>ti \<rho> k'\<rangle> = vars_term (ti \<rho> k') \<union>  set (Z \<rho> k')" by auto
           from exrho(1) have sig_R_r:"sig_R rr" unfolding sig_R_def rho funs_ctrs_def funs_rule_def by force
           from vars_r_subset_tk_Z[OF exrho(1)[unfolded rho] lcs] rho k have 
            vr:"vars_term rr \<subseteq> vars_term (ti \<rho> k') \<union>  set (Z \<rho> k')" by auto 
           from \<theta>s have z_steps:"\<forall>j<length (Z \<rho> k').  ((\<theta>s ! ?k) (Z \<rho> k' ! j), \<sigma>i ?k (Z \<rho> k' ! j)) \<in> par_rstep UR ^^ ?mi ?k" unfolding k by blast
           from \<theta>s have t_steps:"\<forall>x\<in>vars_term (ti \<rho> k'). ((\<theta>s ! ?k) x, \<sigma>i ?k x) \<in> par_rstep UR ^^ ?mi ?k" unfolding k by blast
           with vr z_steps have "\<forall>x\<in>vars_term rr. ((\<theta>s ! ?k) x, \<sigma>i ?k x) \<in> par_rstep UR ^^ ?mi ?k" unfolding set_conv_nth by blast
           with all_ctxt_closed_subst_step[OF acc] have rsteps:"(rr \<cdot> (\<theta>s ! ?k), rr \<cdot> (\<sigma>i ?k)) \<in> par_rstep UR ^^ ?mi ?k" by fast
           from conjunct1[OF conjunct2[OF \<theta>s], rule_format, of "length cs", OF lessI] lcs rho vars_r_subset_tk_Z[OF exrho(1)[unfolded rho] lcs] Uk
            have "\<And>x. x \<in> vars_term rr \<Longrightarrow> sig_R ((\<theta>s ! ?k) x)" unfolding k by auto
           with sig_R_r have 1:"sig_R (rr \<cdot> \<theta>s ! ?k)" unfolding sig_R_def funs_term_subst by blast
           from dec(2)[rule_format, of ?k, unfolded k] rhs_n.simps[of "(l,rr)" cs "Suc k'"] have usq_r:"usq ! Suc (ji (Suc k')) = rr \<cdot> (\<sigma>i  ?k)" 
            unfolding k rho rstep_r_p_s_def ctxt_of_pos_term.simps Let_def ctxt_apply.simps snd_conv by auto
           from ui(3) j(2) have aux:"\<forall>i<Suc m - j. (ui (i + j), ui (Suc i + j)) \<in> par_rstep UR" by simp
           from last_step j_Suc have usq_j:"usq ! j = rr \<cdot> (\<sigma>i ?k)" unfolding rstep_r_p_s_def rho rhs_n.simps by simp
           have "(usq ! j, usq ! (Suc m)) \<in> par_rstep UR ^^ (Suc m - j)" unfolding relpow_fun_conv 
            by (rule exI[of _ "\<lambda> i. ui (i + j)"], insert j(2) aux uk, auto) 
           with ui(2) usq_j have steps:"(rr \<cdot> (\<sigma>i ?k), t \<cdot> \<sigma>) \<in> par_rstep UR ^^ (Suc m - j)" unfolding uk[OF lessI] relpow_add by fastforce
           let ?m = "?mi ?k + (Suc m - j)" 
           from steps rsteps have steps:"(rr \<cdot> (\<theta>s ! ?k), t \<cdot> \<sigma>) \<in> par_rstep UR ^^ ?m" unfolding uk[OF lessI] relpow_add by fastforce
           have "listsum (map (Suc \<circ> ni) [Suc 0..<Suc k']) \<ge>  listsum (map ni [1..<Suc k'])" by (induct k', auto)
           with dec(5) j(2) have m':"?m < m" unfolding k by linarith
           hence m:"((?m, rr \<cdot> (\<theta>s ! ?k)), nt) \<in> measures [fst, size \<circ> snd]" unfolding ns Suc_m by auto
           from 1 tR lint steps ind[rule_format, OF m, unfolded split, rule_format, of t \<sigma>] obtain \<theta> where 
            \<theta>:"?theta_prop_with \<sigma> ?m (rr \<cdot> (\<theta>s ! ?k)) t \<theta> ?k" by blast
           from m' relpow_refl_mono[of "par_rstep UR" ?m n] par_rstep_refl 
             have 1:"\<And> u v. (u,v) \<in> par_rstep UR ^^ ?m \<Longrightarrow>  (u,v) \<in> par_rstep UR ^^ n" unfolding Suc_m by auto
           from \<theta> ssteps have 2:"(s, t\<cdot> \<theta>) \<in> (cstep R)\<^sup>+" by force
           show ?thesis by (rule exI[of _ \<theta>], insert \<theta> 1 2, auto)
         next
         case False 
          let ?s = "Suc m0"
          let ?same = "(usq ! Suc m0, usq ! Suc m) \<in> par_urstep_below_root ^^ (Suc m - Suc m0) "
          let ?dec1 = "\<lambda> m k f. m < k \<and> k < length (snd \<rho>) \<and> (\<exists> ws' Uk. usq ! f = Fun Uk ws' \<and> U \<rho> k = (More Uk Nil Hole (map Var (Z \<rho> k))))"
          let ?step = "\<lambda> ji i \<sigma> . (usq ! ji, usq ! Suc ji) \<in> (rstep_r_p_s UR (lhs_n \<rho> i, rhs_n \<rho> i) \<epsilon> \<sigma> )"
          let ?dec2 = "\<lambda> m k j \<sigma>i f. \<forall>i. (m < i \<and> i \<le> k \<longrightarrow> ?s \<le> j i \<and> j i < f \<and> ?step (j i) i (\<sigma>i i))" 
          let ?dec3 = "\<lambda> m j nm. (usq ! ?s, usq ! (j (Suc m))) \<in>  par_urstep_below_root ^^ nm"
          let ?dec4 = "\<lambda> m k j ni. (\<forall>i. m < i \<and> i < k \<longrightarrow> (usq ! Suc (j i), usq ! (j (Suc i))) \<in> par_urstep_below_root ^^ (ni i))"
          let ?dec5 = "\<lambda> k j nk f. (usq ! Suc (j k), usq ! f) \<in> par_urstep_below_root ^^ nk" 
          let ?dec6 = "\<lambda> m k nm ni nk f. ?s + nm + (listsum (map ni [Suc m..<k])) + (k - m) + nk = f"  
          let ?dec = "\<lambda> m k j nm ni nk \<sigma>i f. (?dec1 m k f \<and> ?dec2 m k j \<sigma>i f \<and> ?dec3 m j nm \<and> ?dec4 m k j ni \<and> ?dec5 k j nk f \<and> ?dec6 m k nm ni nk f)"
          from uk False have nlr:"\<forall>i. Suc m0 \<le> i \<and> i < Suc m \<longrightarrow> \<not> (\<exists>\<tau>. (usq ! i, usq ! Suc i) \<in> rstep_r_p_s UR (lhs_n \<rho> (length cs), rhs_n \<rho> (length cs)) \<epsilon> \<tau>)" by force
          from a lts i have "Suc m0 \<le> Suc m \<and> Suc m < length usq" "0 < length usq" by auto
          from ur_sequence2[OF steps _ exrho(1), unfolded rho snd_conv, OF nlr[unfolded rho] ts1 U0[unfolded rho] lcs this(1) nlv this(2)]
             have "\<exists> j ni nm nk k \<sigma>i. (?same \<or>  ?dec 0 k j nm ni nk \<sigma>i (Suc m))" unfolding rho by auto
          then obtain ji ni n0 nk k' \<sigma>i where cases:"?same \<or> ?dec 0 k' ji n0 ni nk \<sigma>i (Suc m)" by force
          let ?the_k = " \<lambda> k Um ys. (usq ! Suc m = Fun Um ys \<and> U \<rho> k = More Um [] \<box>  (map Var (Z \<rho> k)) \<and> k < length cs)"
          {assume below_root:"?same" 
           with par_rstep_below_root_pow_same_root[OF below_root[unfolded ts1]] U0 lcs have "\<exists>k Um ys.?the_k k Um ys" by auto
          } note A = this 
          {assume a:"\<not>?same"
           with cases have "?dec 0 k' ji n0 ni nk \<sigma>i (Suc m)" by auto
           from conjunct1[OF this] rho a have "\<exists>k Um ys. ?the_k k Um ys" by force
          } with A have  "\<exists>k Um ys. ?the_k k Um ys" by (cases "?same")
          then obtain k Uk ys where Uk:"usq ! Suc m = Fun Uk ys" "U \<rho> k = More Uk [] \<box>  (map Var (Z \<rho> k))" "k < length cs" by blast
          from U_cond[rule_format, OF exrho(1), unfolded rho snd_conv, OF Uk(3)] rho Uk have UkR:"Uk \<notin> funs_ctrs R" by force
          with uk[OF lessI, unfolded ui(2) Uk(1)] tR[unfolded sig_R_def] funs_term_subst[of t \<sigma>] obtain x where tx:"t = Var x" by (cases t, auto)
          from  uk[OF lessI, unfolded ui(2)] have ts:"t \<cdot> \<sigma> = usq ! Suc m" by auto
          from UkR ts Uk tx have 4:"\<not> sig_R (\<sigma> x)" unfolding sig_R_def by force
          let ?\<theta> = "\<lambda> x. l \<cdot> \<theta>1"     
          from theta1(3) all_ctxt_closed_subst_step[OF acc] have lsteps:"(l \<cdot> \<theta>1, l \<cdot> \<sigma>1) \<in> par_rstep UR ^^ m0" by simp
          from i uk have x:"l \<cdot> \<sigma>1 = usq ! m0" by force
          have "(ui m0, ui (Suc m)) \<in> par_rstep UR ^^ (Suc m - m0)" unfolding relpow_fun_conv 
            by (rule exI[of _ "\<lambda> i. ui (i + m0)"], insert i(1) ui(3), auto)
          with uk i(1) have "(usq ! m0, usq ! (Suc m)) \<in> par_rstep UR ^^ (Suc m - m0)" by simp
          with x ts tx have "(l \<cdot> \<sigma>1, \<sigma> x) \<in> par_rstep UR ^^ (Suc m - m0)" by force
          with lsteps i(1) relpow_add[of m0 "Suc m - m0" "par_rstep UR"] 
            have steps:"(t \<cdot> ?\<theta>, \<sigma> x) \<in> par_rstep UR ^^ n" unfolding Suc tx by auto
          from theta1(1) tx have 1:"(s, t \<cdot> ?\<theta>) \<in> (cstep R)\<^sup>*" by force 
          from theta1(2) 3 sig_R_subst have 2:"sig_R_subst ?\<theta> (vars_term t)" unfolding tx by blast 
          show ?thesis by (rule exI[of _ ?\<theta>], insert 1 2 steps 4, simp add: tx)
         qed
        qed
       qed
     qed
   qed
 qed 
 with s t seq show ?thesis by blast
qed

lemma Theorem_4_3:
 assumes ll:"\<forall> \<rho> \<in> R. ultra_rule left_linear_trs \<rho>"
 and nlv:non_LV
 and sig_R_s:"sig_R s"
 and sig_R_t:"sig_R t" 
 and seq:"(s,t) \<in> (rstep UR)^*"
 shows "(s, t) \<in> (cstep R)^*"
proof- 
 fix x :: 'v
 let ?\<sigma> = "\<lambda> y. t"
 have sig_R_x:"sig_R (Var x)" unfolding sig_R_def by auto
 have lin:"linear_term (Var x)" by auto 
 from seq rtrancl_mono[OF rstep_par_rstep, of UR] have "(s, t) \<in> (par_rstep UR)\<^sup>*" by auto
 with rtrancl_imp_relpow obtain n where "(s,(Var x)\<cdot> ?\<sigma>) \<in> (par_rstep UR) ^^ n" by auto
 from Lemma_4_2[OF ll nlv sig_R_s sig_R_x lin this] sig_R_t show ?thesis by force
qed

end

end

